/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.sample.device;

import java.io.IOException;
import java.lang.annotation.Retention;
import java.util.Random;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ExecutorService;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.joylink5.services.iot.client.JoylinkDevice;
import com.joylink5.services.iot.client.JoylinkDeviceErrorCode;
import com.joylink5.services.iot.client.JoylinkDeviceFunction;
import com.joylink5.services.iot.client.JoylinkDeviceProperty;
import com.joylink5.services.iot.client.JoylinkOtaParameter;
import com.joylink5.services.iot.client.JoylinkOtaParameter.OtaSigAlgorithm;
import com.joylink5.services.iot.client.JoylinkOtaParameter.OtaErrorCode;
import com.joylink5.services.iot.client.JoylinkOtaParameter.OtaState;
import com.joylink5.services.iot.client.JoylinkOtaFunctions;


/**
 * This class encapsulates an actual device. It extends {@link JoylinkDevice} to
 * define properties that are to be kept in sync with device shadow.
 */
public class ConnectedWindow extends JoylinkDevice implements JoylinkOtaFunctions {

    // members for ota task
    private Thread otaTaskThread;

    private String otaCurrentVersion;
    private String otaTaskId;
    private String otaVersion;
    private String otaPackageName;
    private String otaUrl;
    private String otaErrorCode;
    private String otaState;
    private int    otaProgress;

    public ConnectedWindow(String deviceId) {
        super(deviceId);

        super.setOtaFunctionsListener(this);

        // intialise
        initialise();
    }

    // Device initialise
    public void initialise() {
        otaCurrentVersion = "v1.0.0";
    }

    @JoylinkDeviceProperty (name = "window.roomTemperature")
    private float roomTemperature;

    @JoylinkDeviceProperty (name = "window.windowOpen")
    private boolean windowOpen;

    public boolean getWindowOpen() {
        // 1. read the window state from the window actuator
        boolean reportedState = this.windowOpen;
        System.out.println(
                System.currentTimeMillis() + " >>> reported window_windowOpen state: " + (reportedState ? "open" : "closed"));

        
        // 2. return the current window state
        return reportedState;
    }

    public void setWindowOpen(boolean desiredState) {
        // 1. update the window actuator with the desired state
        System.out.println("window.windowOpen should be set:" + String.valueOf(desiredState));
        this.windowOpen = desiredState;

        System.out.println(
                System.currentTimeMillis() + " <<< desired window state to " + (this.windowOpen ? "open" : "closed"));
    }

    public float getRoomTemperature() {
        // 1. Read the actual room temperature from the thermostat
        Random rand = new Random();
        float minTemperature = 20.0f;
        float maxTemperature = 85.0f;
        float reportedTemperature = rand.nextFloat() * (maxTemperature - minTemperature) + minTemperature;

        // 2. (optionally) update the local copy
        this.roomTemperature = reportedTemperature;

        // 3. return the current room temperature
        System.out.println(System.currentTimeMillis() + " >>> reported room temperature: " + reportedTemperature);
        return this.roomTemperature;
    }

    public void setRoomTemperature(float desiredTemperature) {
        // no-op as room temperature is a read-only property. It's not required
        // to have this setter.
    }

    @JoylinkDeviceFunction (name = "window.open")
    private String openWindow;
    public String getOpenWindow() {
        return this.openWindow;
    }

    public void setOpenWindow(String inParam) {
        System.out.println("Execute Function Call window.open With Parametres:" + inParam);
        this.openWindow = "{\"message\":\"success\"}";
    }

    /***************************************************************************************
     * Device OTA
     ***************************************************************************************/
    public String getOtaCurrentVersion() {
        return this.otaCurrentVersion;
    }

    @Override
    public void onFuncOta(String argsJson) {
        System.out.println("\r\n\r\nHello World From onFuncOta \r\n\r\n");

        // get task-id url version
        ObjectMapper mapper = new ObjectMapper();
        JsonNode rootNode;
        try {
            rootNode = mapper.readTree(argsJson);
            this.otaTaskId = rootNode.get("task-id").asText();
            this.otaVersion = rootNode.get("version").asText();
            this.otaUrl = rootNode.get("url").asText();
            this.otaPackageName = rootNode.get("package-name").asText();
        } catch (IOException e) {
            System.out.println("Parse OTA Function Call ota.ota in parameter failed");
            super.setErrorCode(JoylinkDeviceErrorCode.BAD_REQUEST);
            return;
        }

        // execute thread for ota task
        otaTaskThread = new Thread(){
            @Override
            public void run() {
                super.run();
                System.out.println("Start ota task:" + otaTaskId + " procedure...");
                // Intialise State
                otaProgress = 0;
                otaErrorCode = OtaErrorCode.NONE.getValue();
                otaState = OtaState.PREPARE.getValue();

                try {
                    Thread.sleep(2000);

                    // send state change report
                    sendOtaEventStateChanged(otaTaskId, otaState, otaErrorCode, otaProgress);
                    Thread.sleep(1000);

                    // according to install-mode, decided if should interactive with user

                    // start to download package according to url
                    // do download procedure
                    otaState = OtaState.DOWNLOADING.getValue();
                    otaProgress = 1;
                    System.out.println("\r\nDownloading package from " + otaUrl + "\r\n");
                    sendOtaEventStateChanged(otaTaskId, otaState, otaErrorCode, otaProgress);
                    Thread.sleep(1000);

                    // when download done, send state changed
                    otaState = OtaState.DOWNLOAD_DONE.getValue();
                    otaProgress = 10;
                    sendOtaEventStateChanged(otaTaskId, otaState, otaErrorCode, otaProgress);
                    Thread.sleep(1000);

                    // start intall procedure
                    otaState = OtaState.INSTALLING.getValue();
                    otaProgress = 10;
                    sendOtaEventStateChanged(otaTaskId, otaState, otaErrorCode, otaProgress);
                    Thread.sleep(1000);

                    // 
                    otaProgress = 80;
                    sendOtaEventStateChanged(otaTaskId, otaState, otaErrorCode, otaProgress);
                    Thread.sleep(1000);

                    // Install Completed
                    otaState = OtaState.SUCCESS.getValue();
                    otaProgress = 100;
                    sendOtaEventStateChanged(otaTaskId, otaState, otaErrorCode, otaProgress);
                    Thread.sleep(100);

                    // send Version changed
                    otaCurrentVersion = otaVersion;
                    sendOtaEventVersionChanged(otaCurrentVersion);

                } catch (InterruptedException e) {
                    System.out.println("\r\n\r\nOta Thread Recv interrupt\r\n\r\n");
                }
            }
        };
        otaTaskThread.start();

    }

    @Override
    public void onFuncCancelTask(String taskId) {
        System.out.println("\r\n\r\nCancel OTA Task:" + taskId + "\r\n\r\n\r\n");
        if (otaTaskThread.isAlive()) {
            otaTaskThread.interrupt();
            otaState = OtaState.FAILED.getValue();
            otaErrorCode = OtaErrorCode.TASK_CANCEL.getValue();
        }
    }

    @Override
    public String onFuncQueryState(String taskId) {
        System.out.println("\r\n\r\nQuery State OTA Task:" + taskId + "\r\n\r\n\r\n");
        return super.onOtaBuildState(otaTaskId, otaState, otaErrorCode, otaProgress);
    }

    @Override
    public void onFuncExecTask(String taskId) {
        System.out.println("\r\n\r\nCancel OTA Task:" + taskId + "\r\n\r\n\r\n");
    }

}
