/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.sample.device;

import javax.net.ssl.SSLSocketFactory;

import com.joylink5.services.iot.client.JoylinkConnectionStatus;
import com.joylink5.services.iot.client.JoylinkDevice;
import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkMqttClient;
import com.joylink5.services.iot.client.JoylinkQos;
import com.joylink5.services.iot.client.sample.sampleUtil.CommandArguments;
import com.joylink5.services.iot.client.sample.sampleUtil.SampleUtil;
import com.joylink5.services.iot.client.sample.sampleUtil.SampleUtil.KeyStorePasswordPair;

/**
 * This example demonstrates how to use {@link JoylinkMqttClient} and a derived
 * {@link JoylinkDevice} to keep the device connected with its shadow in the
 * cloud. This is the recommended way of accessing a shadow and keeping it
 * synchronized with the device. The only thing the application needs to provide
 * in the derived class is the annotated device properties and their
 * corresponding getter and setter methods.
 */
public class DeviceSample {

    private static JoylinkMqttClient joylinkClient;

    public static void setClient(JoylinkMqttClient client) {
        joylinkClient = client;
    }

    private static void initClient(CommandArguments arguments) {
        String clientEndpoint = arguments.getNotNull("clientEndpoint", SampleUtil.getConfig("clientEndpoint"));
        String deviceId = arguments.getNotNull("deviceId", SampleUtil.getConfig("deviceId"));

        String caCertificateFile = arguments.get("caCertificateFile", SampleUtil.getConfig("caCertificateFile"));
        System.out.println("caCertificateFile:" + caCertificateFile);
        String certificateFile = arguments.get("certificateFile", SampleUtil.getConfig("certificateFile"));
        System.out.println("certificateFile:" + certificateFile);
        String privateKeyFile = arguments.get("privateKeyFile", SampleUtil.getConfig("privateKeyFile"));
        System.out.println("privateKeyFile:" + privateKeyFile);
        int clientEndpointPort = Integer.valueOf(arguments.get("clientEndpointPort", SampleUtil.getConfig("clientEndpointPort")));
        boolean enableGzip = Boolean.parseBoolean(arguments.get("enableGzip", SampleUtil.getConfig("enableGzip")));

        if (joylinkClient == null && certificateFile != null && privateKeyFile != null) {
            String algorithm = arguments.get("keyAlgorithm", SampleUtil.getConfig("keyAlgorithm"));

            // KeyStorePasswordPair pair = SampleUtil.getKeyStorePasswordPair(caCertificateFile, certificateFile, privateKeyFile, algorithm);

            // joylinkClient = new JoylinkMqttClient(clientEndpoint, clientId, pair.keyStore, pair.keyPassword);
            SSLSocketFactory sslSocketFactory = null;
            try{
                sslSocketFactory = SampleUtil.getSocketFactory(caCertificateFile, certificateFile, privateKeyFile, "", algorithm);
            }
            catch(Exception e){
                System.out.println("Create ssl socket factory failed.");
            }
            joylinkClient = new JoylinkMqttClient(clientEndpoint, deviceId, sslSocketFactory, clientEndpointPort,enableGzip);

        }

        if (joylinkClient == null) {
            joylinkClient = new JoylinkMqttClient(clientEndpoint, deviceId, clientEndpointPort,enableGzip);
        }

        if (joylinkClient == null) {
            String iotAccessKeyId = arguments.get("iotAccessKeyId", SampleUtil.getConfig("iotAccessKeyId"));
            String iotSecretAccessKey = arguments.get("iotSecretAccessKey", SampleUtil.getConfig("iotSecretAccessKey"));
            String sessionToken = arguments.get("sessionToken", SampleUtil.getConfig("sessionToken"));

            if (iotAccessKeyId != null && iotSecretAccessKey != null) {
                joylinkClient = new JoylinkMqttClient(clientEndpoint, deviceId, iotAccessKeyId, iotSecretAccessKey,
                        sessionToken,enableGzip);
            }
        }

        if (joylinkClient == null) {
            throw new IllegalArgumentException("Failed to construct client due to missing certificate or credentials.");
        }
    }

    public static void main(String args[]) throws InterruptedException, JoylinkException {
        System.out.println("breakpoint");
        CommandArguments arguments = CommandArguments.parse(args);
        initClient(arguments);

        joylinkClient.setWillMessage(new JoylinkMessage("client/disconnect", JoylinkQos.QOS0, joylinkClient.getClientId()));

        String deviceId = arguments.getNotNull("deviceId", SampleUtil.getConfig("deviceId"));
        ConnectedWindow connectedWindow = new ConnectedWindow(deviceId);
        connectedWindow.setThingModelInfo("urn:user-spec-v1:thing-model:001:9545fe83892112f08bf4cfadf8f51890", "v1.0.0");
        joylinkClient.attach(connectedWindow);


        joylinkClient.connect();

        // main task loop
        long lastReportTime = 0;
        long timeCurrent;
        String propJson;

        JoylinkConnectionStatus newStatus;
        JoylinkConnectionStatus status = JoylinkConnectionStatus.DISCONNECTED;
        while (true) {
            newStatus = joylinkClient.getConnectionStatus();
            if (!status.equals(newStatus)) {
                System.out.println(System.currentTimeMillis() + " Connection status changed to " + newStatus);
                status = newStatus;

                // when connect do some action which are needed
                if (newStatus.equals(JoylinkConnectionStatus.CONNECTED)) {
                    // Send thing-model post
                    connectedWindow.sendThingModelPost();

                    // Send properties post
                    Thread.sleep(1000);
                    propJson = connectedWindow.onPropertiesPost();
                    connectedWindow.sendPropertiesPost(connectedWindow.getLocalVersion().get(), propJson);

                    // Send ota version
                    Thread.sleep(1000);
                    connectedWindow.sendOtaEventVersionChanged(connectedWindow.getOtaCurrentVersion());
                }    
            }


            if (newStatus.equals(JoylinkConnectionStatus.CONNECTED)) {
                // send property post and event post
                timeCurrent = System.currentTimeMillis();
                if ((timeCurrent - lastReportTime) > 20000) {
                    // send properties post
                    propJson = connectedWindow.onPropertiesPost();
                    connectedWindow.sendPropertiesPost(connectedWindow.getLocalVersion().get(), propJson);

                    // send event, actully event should be trigged under some conditions
                    if (true) {
                        String evtKey = "window.alter";
                        String paraJson = "{\"message\":\"danger, danger,danger!!\"}";
                        connectedWindow.sendEventPost(evtKey, paraJson);
                    }

                    lastReportTime = timeCurrent;
                }

            }

            Thread.sleep(1000);
        }
    }

}
