/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.sample.gateway;

import javax.net.ssl.SSLSocketFactory;

import com.joylink5.services.iot.client.JoylinkConnectionStatus;
import com.joylink5.services.iot.client.JoylinkDevice;
import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkMqttClient;
import com.joylink5.services.iot.client.JoylinkQos;
import com.joylink5.services.iot.client.sample.sampleUtil.CommandArguments;
import com.joylink5.services.iot.client.sample.sampleUtil.SampleUtil;
import com.joylink5.services.iot.client.sample.sampleUtil.SampleUtil.KeyStorePasswordPair;

/**
 * This example demonstrates how to use {@link JoylinkMqttClient} and a derived
 * {@link JoylinkDevice} to keep the device connected with its shadow in the
 * cloud. This is the recommended way of accessing a shadow and keeping it
 * synchronized with the device. The only thing the application needs to provide
 * in the derived class is the annotated device properties and their
 * corresponding getter and setter methods.
 */
public class GatewaySample {

    private static JoylinkMqttClient joylinkClient;

    public static void setClient(JoylinkMqttClient client) {
        joylinkClient = client;
    }

    private static void initClient(CommandArguments arguments) {
        String clientEndpoint = arguments.getNotNull("clientEndpoint", SampleUtil.getConfig("clientEndpoint"));
        String deviceId = arguments.getNotNull("deviceId", SampleUtil.getConfig("deviceId"));

        String caCertificateFile = arguments.get("caCertificateFile", SampleUtil.getConfig("caCertificateFile"));
        String certificateFile = arguments.get("certificateFile", SampleUtil.getConfig("certificateFile"));
        String privateKeyFile = arguments.get("privateKeyFile", SampleUtil.getConfig("privateKeyFile"));
        int clientEndpointPort = Integer.valueOf(arguments.get("clientEndpointPort", SampleUtil.getConfig("clientEndpointPort")));
        boolean enableGzip = Boolean.parseBoolean(arguments.get("enableGzip", SampleUtil.getConfig("enableGzip")));

        if (joylinkClient == null && certificateFile != null && privateKeyFile != null) {
            String algorithm = arguments.get("keyAlgorithm", SampleUtil.getConfig("keyAlgorithm"));

            // KeyStorePasswordPair pair = SampleUtil.getKeyStorePasswordPair(caCertificateFile, certificateFile, privateKeyFile, algorithm);

            // joylinkClient = new JoylinkMqttClient(clientEndpoint, clientId, pair.keyStore, pair.keyPassword);
            SSLSocketFactory sslSocketFactory = null;
            try{
                sslSocketFactory = SampleUtil.getSocketFactory(caCertificateFile, certificateFile, privateKeyFile, "", algorithm);
            }
            catch(Exception e){
                System.out.println("Create ssl socket factory failed.");
            }
            joylinkClient = new JoylinkMqttClient(clientEndpoint, deviceId, sslSocketFactory, clientEndpointPort,enableGzip);

        }

        if (joylinkClient == null) {
            joylinkClient = new JoylinkMqttClient(clientEndpoint, deviceId, clientEndpointPort,enableGzip);
        }

        if (joylinkClient == null) {
            String iotAccessKeyId = arguments.get("iotAccessKeyId", SampleUtil.getConfig("iotAccessKeyId"));
            String iotSecretAccessKey = arguments.get("iotSecretAccessKey", SampleUtil.getConfig("iotSecretAccessKey"));
            String sessionToken = arguments.get("sessionToken", SampleUtil.getConfig("sessionToken"));

            if (iotAccessKeyId != null && iotSecretAccessKey != null) {
                joylinkClient = new JoylinkMqttClient(clientEndpoint, deviceId, iotAccessKeyId, iotSecretAccessKey,
                        sessionToken,enableGzip);
            }
        }

        if (joylinkClient == null) {
            throw new IllegalArgumentException("Failed to construct client due to missing certificate or credentials.");
        }
    }

    public static void main(String args[]) throws InterruptedException, JoylinkException {
        CommandArguments arguments = CommandArguments.parse(args);
        initClient(arguments);

        joylinkClient.setWillMessage(new JoylinkMessage("client/disconnect", JoylinkQos.QOS0, joylinkClient.getClientId()));

        String deviceId = arguments.getNotNull("deviceId", SampleUtil.getConfig("deviceId"));
        ConnectedWindow gateWay = new ConnectedWindow(deviceId);
        gateWay.setThingModelInfo("urn:user-spec-v1:thing-model:002:b62c09f383f9adf5c0fffbd103698e58", "v1.0.0");
        joylinkClient.attach(gateWay);

        SubWindow subWindow = new SubWindow("10106e45334b300");
        subWindow.setSubDevInfo("java_sdk_subdevice", "10106e45334b300", "", "urn:user-spec-v1:thing-model:003:b8fd0d04580df40839663b02ea810147", "", "AABBCCDD", "tcp", "");
        joylinkClient.attach(subWindow);


        joylinkClient.connect();


        // task loop
        String propJson;
        long lastReportTime = 0;
        long timeCurrent;
        JoylinkConnectionStatus newStatus;
        JoylinkConnectionStatus status = JoylinkConnectionStatus.DISCONNECTED;
        while (true) {
            newStatus = joylinkClient.getConnectionStatus();
            if (!status.equals(newStatus)) {
                System.out.println(System.currentTimeMillis() + " Connection status changed to " + newStatus);
                status = newStatus;

                // when connected do some actions which are needed
                if (newStatus.equals(JoylinkConnectionStatus.CONNECTED)) {
                    // Gateway device send thing-model post
                    gateWay.sendThingModelPost();

                    // Gateway send ota property
                    Thread.sleep(1000);
                    propJson = gateWay.onOtaPropertiesPost();
                    gateWay.sendOtaPropertiesPost(propJson);

                    // Gateway send ota version
                    Thread.sleep(1000);
                    gateWay.sendOtaEventVersionChanged(gateWay.getOtaCurrentVersion());

                    // Subdevice Send Event add-devices
                    // Thread.sleep(1000);
                    // subWindow.sendEventAddDevices();

                    // Subdevice Send online state
                    Thread.sleep(1000);
                    subWindow.sendDeviceOnlineState(true);

                }
            }

            // subdevice property post and event post
/*
            timeCurrent = System.currentTimeMillis();
            if (newStatus.equals(JoylinkConnectionStatus.CONNECTED)) {
                if ((timeCurrent - lastReportTime) > 20000) {
                    propJson = subWindow.onPropertiesPost();
                    subWindow.sendPropertiesPost(subWindow.getLocalVersion().get(), propJson);

                    // this is just a example, actully, event should be sent under some conditions
                    if (true) {
                        String evtKey = "subwindow.alter";
                        String paraJson = "{\"message\":\"alter, alter, alter!!\"}";
                        subWindow.sendEventPost(evtKey, paraJson);
                    }
                    lastReportTime = timeCurrent;
                }
            }
*/
            Thread.sleep(1000);
        }
    }

}
