/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.sample.pubSub;

import com.joylink5.services.iot.client.JoylinkMqttClient;

import javax.net.ssl.SSLSocketFactory;

import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkQos;
import com.joylink5.services.iot.client.JoylinkTimeoutException;
import com.joylink5.services.iot.client.JoylinkTopic;
import com.joylink5.services.iot.client.sample.sampleUtil.CommandArguments;
import com.joylink5.services.iot.client.sample.sampleUtil.SampleUtil;
import com.joylink5.services.iot.client.sample.sampleUtil.SampleUtil.KeyStorePasswordPair;

/**
 * This is an example that uses {@link JoylinkMqttClient} to subscribe to a topic and
 * publish messages to it. Both blocking and non-blocking publishing are
 * demonstrated in this example.
 */
public class PublishSubscribeSample {

    private static final String TestTopic = "sdk/test/java";
    private static final JoylinkQos TestTopicQos = JoylinkQos.QOS0;

    private static JoylinkMqttClient joylinkClient;

    public static void setClient(JoylinkMqttClient client) {
        joylinkClient = client;
    }

    public static class BlockingPublisher implements Runnable {
        private final JoylinkMqttClient joylinkClient;

        public BlockingPublisher(JoylinkMqttClient joylinkClient) {
            this.joylinkClient = joylinkClient;
        }

        @Override
        public void run() {
            long counter = 1;

            while (true) {
                String payload = "hello from blocking publisher - " + (counter++);
                try {
                    joylinkClient.publish(TestTopic, payload);
                } catch (JoylinkException e) {
                    System.out.println(System.currentTimeMillis() + ": publish failed for " + payload);
                }
                System.out.println(System.currentTimeMillis() + ": >>> " + payload);

                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    System.out.println(System.currentTimeMillis() + ": BlockingPublisher was interrupted");
                    return;
                }
            }
        }
    }

    public static class NonBlockingPublisher implements Runnable {
        private final JoylinkMqttClient joylinkClient;

        public NonBlockingPublisher(JoylinkMqttClient joylinkClient) {
            this.joylinkClient = joylinkClient;
        }

        @Override
        public void run() {
            long counter = 1;

            while (true) {
                String payload = "hello from non-blocking publisher - " + (counter++);
                JoylinkMessage message = new NonBlockingPublishListener(TestTopic, TestTopicQos, payload);
                try {
                    joylinkClient.publish(message);
                } catch (JoylinkException e) {
                    System.out.println(System.currentTimeMillis() + ": publish failed for " + payload);
                }

                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    System.out.println(System.currentTimeMillis() + ": NonBlockingPublisher was interrupted");
                    return;
                }
            }
        }
    }

    private static void initClient(CommandArguments arguments) {
        String clientEndpoint = arguments.getNotNull("clientEndpoint", SampleUtil.getConfig("clientEndpoint"));
        String clientId = arguments.getNotNull("clientId", SampleUtil.getConfig("clientId"));

        String caCertificateFile = arguments.get("caCertificateFile", SampleUtil.getConfig("caCertificateFile"));
        String certificateFile = arguments.get("certificateFile", SampleUtil.getConfig("certificateFile"));
        String privateKeyFile = arguments.get("privateKeyFile", SampleUtil.getConfig("privateKeyFile"));
        boolean enableGzip = Boolean.parseBoolean(arguments.get("enableGzip", SampleUtil.getConfig("enableGzip")));


        if (joylinkClient == null && certificateFile != null && privateKeyFile != null) {
            String algorithm = arguments.get("keyAlgorithm", SampleUtil.getConfig("keyAlgorithm"));

            // KeyStorePasswordPair pair = SampleUtil.getKeyStorePasswordPair(caCertificateFile, certificateFile, privateKeyFile, algorithm);

            // joylinkClient = new JoylinkMqttClient(clientEndpoint, clientId, pair.keyStore, pair.keyPassword);
            SSLSocketFactory sslSocketFactory = null;
            try{
                sslSocketFactory = SampleUtil.getSocketFactory(caCertificateFile, certificateFile, privateKeyFile, "", algorithm);
            }
            catch(Exception e){
                System.out.println("Create ssl socket factory failed.");
            }
            joylinkClient = new JoylinkMqttClient(clientEndpoint, clientId, sslSocketFactory, 8883,enableGzip);

        }

        if (joylinkClient == null) {
            String iotAccessKeyId = arguments.get("iotAccessKeyId", SampleUtil.getConfig("iotAccessKeyId"));
            String iotSecretAccessKey = arguments.get("iotSecretAccessKey", SampleUtil.getConfig("iotSecretAccessKey"));
            String sessionToken = arguments.get("sessionToken", SampleUtil.getConfig("sessionToken"));

            if (iotAccessKeyId != null && iotSecretAccessKey != null) {
                joylinkClient = new JoylinkMqttClient(clientEndpoint, clientId, iotAccessKeyId, iotSecretAccessKey,
                        sessionToken,enableGzip);
            }
        }

        if (joylinkClient == null) {
            throw new IllegalArgumentException("Failed to construct client due to missing certificate or credentials.");
        }
    }

    public static void main(String args[]) throws InterruptedException, JoylinkException, JoylinkTimeoutException {
        CommandArguments arguments = CommandArguments.parse(args);
        initClient(arguments);

        joylinkClient.connect();

        JoylinkTopic topic = new TestTopicListener(TestTopic, TestTopicQos);
        joylinkClient.subscribe(topic, true);

        Thread blockingPublishThread = new Thread(new BlockingPublisher(joylinkClient));
        Thread nonBlockingPublishThread = new Thread(new NonBlockingPublisher(joylinkClient));

        blockingPublishThread.start();
        nonBlockingPublishThread.start();

        blockingPublishThread.join();
        nonBlockingPublishThread.join();
    }

}
