/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.joylink5.services.iot.client.JoylinkConstant;
import com.joylink5.services.iot.client.protocol.AbstractJoylinkDevice;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.Command;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.CommandAck;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.ServiceKey;

import com.joylink5.services.iot.client.JoylinkOtaFunctions;


public class JoylinkDevice extends AbstractJoylinkDevice {

    private static final Logger LOGGER = Logger.getLogger(JoylinkDevice.class.getName());

    protected String thingModelId;
    protected String thingModelVersion;

    /**
     * Instantiates a new device instance.
     *
     * @param deviceId
     *          the device id
     */
    public JoylinkDevice(String deviceId) {
        super(deviceId);
    }

    /**
     * This function handles thing-model info Set.
     *
     * @param id
     *          thing-model id
     * @param version
     *          thing-model version
     */
    public void setThingModelInfo(String id, String version) {
        this.thingModelId = id;
        this.thingModelVersion = version;
    }

    /**
     * Handles device activate, this function will be called when client connect
     * successfully, device atech.
     *
     * @return the report interval in milliseconds.
     */
    @Override
    public void activate() throws JoylinkException {
        super.activate();
        LOGGER.info("Enter Device Activate");
    }

    /**
     * Handles device deactivate, this function will be called when client connect
     * failed, connection lost and device detach.
     *
     * @return the report interval in milliseconds.
     */
    @Override
    public void deactivate() throws JoylinkException {
        super.deactivate();
        LOGGER.info("Enter Device Deactivate");
    }

    /**
     * Checks if version is enabled for device properties.
     *
     * @return true, if versioning is enabled.
     */
    @Override
    public boolean isEnableVersioning() {
        return super.isEnableVersioning();
    }

    /**
     * Sets the device properties' version to be enabled or disabled. This value
     * must be set before the device is attached to a client via the
     * {@link JoylinkMqttClient#attach(JoylinkDevice)} call.
     *
     * @param enableVersioning
     *          true to enable device properties' version; false to disable.
     */
    @Override
    public void setEnableVersioning(boolean enableVersioning) {
        super.setEnableVersioning(enableVersioning);
    }

    /**
     * This function handles sending message of thing-model post.
     *
     */
    public void sendThingModelPost() {
        if ((this.thingModelId == null) || (this.thingModelVersion == null)) {
            LOGGER.severe("[ERROR] Send Thing Model is Failed For Information is not completed!");
            return;
        }

        // Build Thing Model Information
        StringBuilder thingModel = new  StringBuilder("{");
        thingModel.append("\"").append(JoylinkConstant.PAYLOAD_THMD_ID).append("\":\"");
        thingModel.append(this.thingModelId).append("\",");

        thingModel.append("\"").append(JoylinkConstant.PAYLOAD_THMD_VER).append("\":\"");
        thingModel.append(this.thingModelVersion).append("\"}");

        // Build Payload
        StringBuilder payload = new StringBuilder("{");
        payload.append("\"").append(JoylinkConstant.PAYLOAD_DEV_ID).append("\":\"");
        payload.append(this.deviceId).append("\",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_TIME_STAMP).append("\":");
        payload.append(System.currentTimeMillis()/1000).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_THMD).append("\":");
        payload.append(thingModel.toString()).append("}");

        JoylinkQos qos = JoylinkQos.valueOf(JoylinkConfig.DEVICE_THING_MODEL_POST_QOS);
        JoylinkMessage message = new JoylinkMessage(null, qos, payload.toString());

        try {
            super.getCommandManager().runCommand(Command.THMD_POST, message, super.getClient().getServerAckTimeout(), true);
        } catch (JoylinkTimeoutException e) {
            // async command, shouldn't receive timeout exception
        } catch (JoylinkException e) {
            LOGGER.log(Level.WARNING, "Failed to publish device report message", e);
        }
    }

    /**
     * This function handles collecting properties set for poroperties post.
     * The default implementation gets properties that are defined by 
     * Annotation class {JoylinkDeviceProperty}.
     *
     * One could override this default implementation.
     *
     * @return the JSON document containing property
     *         like this: {"window.windowOpen":false,"window.roomTemperature":21.462763}
     */
    @Override
    public String onPropertiesPost() {
        return super.onPropertiesPost();
    }

    /**
     * This function handles sending message of properties post.
     *
     * @param reportVersion
     *          current propeerties' version
     *
     * @param jsonProp
     *          the JSON document containing property
     *          like this: {"window.windowOpen":false,"window.roomTemperature":21.462763}
     * 
     */
    @Override
    public void sendPropertiesPost(long reportVersion, String jsonProp) {
        super.sendPropertiesPost(null, reportVersion, jsonProp);
    }

    /**
     * This function handles device operation of properties set.
     * The default implementation is to find and set properties
     * that are defined by Annotation class {JoylinkDeviceProperty}.
     *
     * One could override this default implementation.
     *
     * @param jsonProp
     *          the JSON document containing property and value that are needed to be set
     *          like this: {"window.windowOpen":false}
     * @return the JSON document containing property after operation 'properties set'
     *          like this: {"window.windowOpen":false}
     */
    @Override
    public String onPropertiesSet(String jsonProp) {
        return super.onPropertiesSet(jsonProp);
    }

    /**
     * This function handles sending message of properties set response.
     *
     * @param messageId
     *          the messageId of corresponding request message
     * @param jsonProp
     *          the JSON document cotaining property and value after operation 'propeties set'
     * @param errCode
     *          error code, definition refers to {JoylinkDeviceErrorCode}
     */
    @Override
    public void sendPropertiesSetResponse(String messageId, String jsonProp, long errCode) {
        super.sendPropertiesSetResponse(null, messageId, jsonProp, errCode);
    }

    /**
     * This function handles device operation of properties get.
     * The default implementation is to find and get properties
     * that are defined by Annotation class {JoylinkDeviceProperty}.
     *
     * One could override this default implementation.
     *
     * @param jsonProp
     *          the JSON document containing property array
     *          like this: ["window.windowOpen"]
     * @return the JSON document containing prooperty after operation 'properties get'
     *          like this: {"window.windowOpen":false}
     */
    @Override
    public String onPropertiesGet(String jsonProp) {
        return super.onPropertiesGet(jsonProp);
    }

    /**
     * This function handles sending message of properties get response.
     *
     * @param messageId
     *          the messageId of corresponding request message
     * @param jsonProp
     *          the JSON document cotaining property and value after operation 'propeties get'
     * @param errCode
     *          error code, definition refers to {JoylinkDeviceErrorCode}
     */
    @Override
    public void sendPropertiesGetResponse(String messageId, String jsonProp, long errCode) {
        super.sendPropertiesGetResponse(null, messageId, jsonProp, errCode);
    }

    /**
     * This function handles device operation of functions call.
     * The default implementation is to find and get functions
     * that are defined by Annotation class {JoylinkDeviceFunction}.
     *
     * Do not include functions of OTA service and connection agent service.
     *
     * One could override this default implementation.
     *
     * @param key
     *          key of the function call
     * @param inParam
     *          the input paraments of function call
     *          like this: {"inparam":"hello_world"}
     * @return the JSON document containing out parameters
     *          like this: {"message":"success"}
     */
    @Override
    public String onFunctionsCall(String key, String inParam) {
        return super.onFunctionsCall(key, inParam);
    }

    /**
     * This function handles sending message of functions call response.
     *
     * @param messageId
     *          the messageId of corresponding request message
     * @param funcKey
     *          the key of corresponding request function call
     * @param outJson
     *          the JSON document cotaining out parametes
     * @param errCode
     *          error code, definition refers to {JoylinkDeviceErrorCode} 
     */
    @Override
    public void sendFunctionCallResponse(String messageId, String funcKey, String outJson, long errCode) {
        super.sendFunctionCallResponse(null, messageId, funcKey, outJson, errCode);
    }

    /**
     * This function handles sending message of device event post.
     *
     * @param eventKey
     *          the key of the event
     * @param paraJson
     *          the JSON document containing parameters of this event
     *
     */
    public void sendEventPost(String eventKey, String paraJson) {
        super.sendEventPost(null, eventKey, paraJson);
    }

    /**
     * This function handles collecting properties set of OTA service.
     *
     * @return the JSON document containing property of OTA service
     */
    @Override
    public String onOtaPropertiesPost() {
        return super.onOtaPropertiesPost();
    }

    /**
     * This function handles sending message of OTA properties post.
     *
     * @param jsonProp
     *          the JSON document containing property of OTA service
     *
     */
    public void sendOtaPropertiesPost(String jsonProp) {
        super.sendPropertiesPost(ServiceKey.OTA, 0, jsonProp);
    }
}
