/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;

/**
 * Annotation class that is used to annotate properties in {@link JoylinkDevice}.
 * Properties annotated must be accessible via corresponding getter and setter
 * methods.
 * <p>
 * With optional values provided by this annotation class, properties can also
 * be configured to disable reporting to the platform or not to accept set/get
 * from the platform. If {@link #enableReport()} is disabled, property getter
 * function is not required. Likewise, if {@link #allowUpdate()} is disabled for
 * a property, its setter method is not required.
 * </p>
 */
@Retention(RetentionPolicy.RUNTIME)
public @interface JoylinkDeviceProperty {

    /**
     * An optional name can be provided to the annotated property.
     * If not provided, the actual property name will be used.
     *
     * @return the name of the property.
     */
    String name() default "";

    /**
     * Enable reporting the annotated property to the platform. It's enabled by
     * default.
     *
     * @return true to enable reporting to the platform, false otherwise.
     */
    boolean enableReport() default true;

    /**
     * Allow updates from the platform. It's enabled by default.
     *
     * @return true to allow get/set from the platform, false otherwise.
     */
    boolean allowUpdate() default true;

}
