/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client;

import com.joylink5.services.iot.client.core.AbstractJoylinkClient;

import javax.net.ssl.SSLSocketFactory;
import java.security.KeyStore;


public class JoylinkMqttClient extends AbstractJoylinkClient {

    /**
     * Instantiates a new client using TLS 1.2 mutual authentication. Client
     * certificate and private key are passed in through the {@link KeyStore}
     * argument. The key password protecting the private key in the
     * {@link KeyStore} is also required.
     *
     * @param clientEndpoint
     *            the MQTT broker address
     * @param clientId
     *            the client ID uniquely identify a MQTT connection. Two clients
     *            with the same client ID are not allowed to be connected
     *            concurrently to a same endpoint.
     * @param keyStore
     *            the key store containing the client X.509 certificate and
     *            private key. 
     * @param keyPassword
     *            the key password protecting the private key in the
     *            {@code keyStore} argument.
     */
    public JoylinkMqttClient(String clientEndpoint, String clientId, KeyStore keyStore, String keyPassword) {
        super(clientEndpoint, clientId, keyStore, keyPassword);
    }

    /**
     * Instantiates a new client using TLS 1.2 mutual authentication. Client
     * certificate and private key should be used to initialize the KeyManager
     * of the socketFactory.
     *
     * @param clientEndpoint
     *            the MQTT broker address
     * @param clientId
     *            the client ID uniquely identify a MQTT connection. Two clients
     *            with the same client ID are not allowed to be connected
     *            concurrently to a same endpoint.
     * @param socketFactory
     *            A socketFactory instantiated with a Keystore containing the client X.509
     *            certificate and private key, and a Truststore containing trusted
     *            Certificate Authorities(CAs).
     */
    public JoylinkMqttClient(String clientEndpoint, String clientId, SSLSocketFactory socketFactory) {
        super(clientEndpoint, clientId, socketFactory);
    }

    /**
     * Instantiates a new client using TLS 1.2 mutual authentication. Client
     * certificate and private key should be used to initialize the KeyManager
     * of the socketFactory.
     *
     * @param clientEndpoint
     *            the MQTT broker address
     * @param clientId
     *            the client ID uniquely identify a MQTT connection. Two clients
     *            with the same client ID are not allowed to be connected
     *            concurrently to a same endpoint.
     * @param socketFactory
     *            A socketFactory instantiated with a Keystore containing the client X.509
     *            certificate and private key, and a Truststore containing trusted
     *            Certificate Authorities(CAs).
     * @param port
     *            The socket port to use.
     * @param enableGzip
     *            is enable gzip
     */
    public JoylinkMqttClient(String clientEndpoint, String clientId, SSLSocketFactory socketFactory, int port,Boolean enableGzip) {
        super(clientEndpoint, clientId, socketFactory, port,true,enableGzip);
    }

    /**
     * Instantiates a new client using TCP.
     *
     * @param clientEndpoint
     *            the MQTT broker address
     * @param clientId
     *            the client ID uniquely identify a MQTT connection. Two clients
     *            with the same client ID are not allowed to be connected
     *            concurrently to a same endpoint.
     * @param port
     *            The socket port to use.
     * @param enableGzip
     *            is enable gzip
     */
    public JoylinkMqttClient(String clientEndpoint, String clientId, int port,Boolean enableGzip) {
        super(clientEndpoint, clientId, port,true,enableGzip);
    }

    /**
     * Instantiates a new client using Secure WebSocket.
     *
     * @param clientEndpoint
     *            the MQTT broker address
     * @param clientId
     *            the client ID uniquely identify a MQTT connection. Two clients
     *            with the same client ID are not allowed to be connected
     *            concurrently to a same endpoint.
     * @param iotAccessKeyId
     *            the access key id
     * @param iotSecretAccessKey
     *            the secret access key
     * @param enableGzip
     *            is enable gzip
     */
    public JoylinkMqttClient(String clientEndpoint, String clientId, String iotAccessKeyId, String iotSecretAccessKey,Boolean enableGzip) {
        super(clientEndpoint, clientId, iotAccessKeyId, iotSecretAccessKey, null,enableGzip);
    }

    /**
     * Instantiates a new client using Secure WebSocket
     *
     * @param clientEndpoint
     *            the MQTT broker address
     * @param clientId
     *            the client ID uniquely identify a MQTT connection. Two clients
     *            with the same client ID are not allowed to be connected
     *            concurrently to a same endpoint.
     * @param iotAccessKeyId
     *            the access key id
     * @param iotSecretAccessKey
     *            the secret access key
     * @param sessionToken
     *            Session token received along with the temporary credentials
     *            from services like STS server.
     * @param enableGzip
     *            is enable gzip
     */
    public JoylinkMqttClient(String clientEndpoint, String clientId, String iotAccessKeyId, String iotSecretAccessKey,
            String sessionToken,Boolean enableGzip) {
        super(clientEndpoint, clientId, iotAccessKeyId, iotSecretAccessKey, sessionToken,enableGzip);
    }

    /**
     * Instantiates a new client using Secure WebSocket
     *
     * @param clientEndpoint
     *            the MQTT broker address
     * @param clientId
     *            the client ID uniquely identify a MQTT connection. Two clients
     *            with the same client ID are not allowed to be connected
     *            concurrently to a same endpoint.
     * @param iotAccessKeyId
     *            the access key id
     * @param iotSecretAccessKey
     *            the secret access key
     * @param sessionToken
     *            Session token received along with the temporary credentials
     *            from services like STS server.
     * @param region
     *            the region
     */
    public JoylinkMqttClient(String clientEndpoint, String clientId, String iotAccessKeyId, String iotSecretAccessKey,
                            String sessionToken, String region) {
        super(clientEndpoint, clientId, iotAccessKeyId, iotSecretAccessKey, sessionToken, region);
    }

    /**
     * Updates credentials used for signing Secure WebSocket URLs. When temporary
     * credentails used for the WebSocket connection are expired, newer
     * credentails can be supplied through this API to allow new connections to
     * be reestablished using the new credentails.
     *
     * @param iotAccessKeyId
     *            the access key id
     * @param iotSecretAccessKey
     *            the secret access key
     * @param sessionToken
     *            Session token received along with the temporary credentials
     *            from services like STS server..
     */
    @Override
    public void updateCredentials(String iotAccessKeyId, String iotSecretAccessKey, String sessionToken) {
        super.updateCredentials(iotAccessKeyId, iotSecretAccessKey, sessionToken);
    }

    /**
     * Gets the number of client threads currently configured. Each client has
     * their own thread pool, which is used to execute user callback functions
     * as well as any timeout callback functions requested. By default, the
     * thread pool is configured with one execution thread.
     *
     * @return the number of client threads
     */
    @Override
    public int getNumOfClientThreads() {
        return super.getNumOfClientThreads();
    }

    /**
     * Sets a new value for the number of client threads. This value must be set
     * before {@link #connect()} is called.
     *
     * @param numOfClientThreads
     *            the new number of client threads. The default value is 1.
     */
    @Override
    public void setNumOfClientThreads(int numOfClientThreads) {
        super.setNumOfClientThreads(numOfClientThreads);
    }

    /**
     * Gets the connection timeout in milliseconds currently configured.
     * Connection timeout specifies how long the client should wait for the
     * connection to be established with the server. By default, it's 30,000ms.
     *
     * @return the connection timeout
     */
    @Override
    public int getConnectionTimeout() {
        return super.getConnectionTimeout();
    }

    /**
     * Sets a new value in milliseconds for the connection timeout. This value
     * must be set before {@link #connect()} is called.
     *
     * @param connectionTimeout
     *            the new connection timeout. The default value is 30,000ms.
     */
    @Override
    public void setConnectionTimeout(int connectionTimeout) {
        super.setConnectionTimeout(connectionTimeout);
    }

    /**
     * Gets the maximum number of connection retries currently configured.
     * Connections will be automatically retried for the configured maximum
     * times when failing to be established or lost. User disconnect, requested
     * via {@link #disconnect()} will not be retried. By default, it's 5 times.
     * Setting it to 0 will disable the connection retry function.
     *
     * @return the max connection retries
     */
    @Override
    public int getMaxConnectionRetries() {
        return super.getMaxConnectionRetries();
    }

    /**
     * Sets a new value for the maximum connection retries. This value must be
     * set before {@link #connect()} is called. Setting it to 0 will disable the
     * connection retry function.
     *
     * @param maxConnectionRetries
     *            the new max connection retries. The default value is 5.
     */
    @Override
    public void setMaxConnectionRetries(int maxConnectionRetries) {
        super.setMaxConnectionRetries(maxConnectionRetries);
    }

    /**
     * Gets the base retry delay in milliseconds currently configured. For each
     * connection failure, a brief delay has to elapse before the connection is
     * retried. The retry delay is calculated using this simple formula
     * {@code delay = min(baseRetryDelay * pow(2, numRetries), maxRetryDelay)}.
     * By default, the base retry delay is 3,000ms.
     *
     * @return the base retry delay
     */
    @Override
    public int getBaseRetryDelay() {
        return super.getBaseRetryDelay();
    }

    /**
     * Sets a new value in milliseconds for the base retry delay. This value
     * must be set before {@link #connect()} is called.
     *
     * @param baseRetryDelay
     *            the new base retry delay. The default value is 3,000ms.
     */
    @Override
    public void setBaseRetryDelay(int baseRetryDelay) {
        super.setBaseRetryDelay(baseRetryDelay);
    }

    /**
     * Gets the maximum retry delay in milliseconds currently configured. For
     * each connection failure, a brief delay has to elapse before the
     * connection is retried. The retry delay is calculated using this simple
     * formula
     * {@code delay = min(baseRetryDelay * pow(2, numRetries), maxRetryDelay)}.
     * By default, the maximum retry delay is 30,000ms.
     *
     * @return the maximum retry delay
     */
    @Override
    public int getMaxRetryDelay() {
        return super.getMaxRetryDelay();
    }

    /**
     * Sets a new value in milliseconds for the maximum retry delay. This value
     * must be set before {@link #connect()} is called.
     *
     * @param maxRetryDelay
     *            the new max retry delay. The default value is 30,000ms.
     */
    @Override
    public void setMaxRetryDelay(int maxRetryDelay) {
        super.setMaxRetryDelay(maxRetryDelay);
    }

    /**
     * Gets the server acknowledge timeout in milliseconds currently configured.
     * This timeout is used internally by the SDK when subscribing to shadow
     * confirmation topics for get, update, and delete requests. It's also used
     * for re-subscribing to user topics when the connection is retried. For
     * most of the APIs provided in the SDK, the user can specify the timeout as
     * an argument. By default, the server acknowledge timeout is 3,000ms.
     *
     * @return the server acknowledge timeout
     */
    @Override
    public int getServerAckTimeout() {
        return super.getServerAckTimeout();
    }

    /**
     * Sets a new value in milliseconds for the default server acknowledge
     * timeout. This value must be set before {@link #connect()} is called.
     *
     * @param serverAckTimeout
     *            the new server acknowledge timeout. The default value is
     *            3,000ms.
     */
    @Override
    public void setServerAckTimeout(int serverAckTimeout) {
        super.setServerAckTimeout(serverAckTimeout);
    }

    /**
     * Gets the keep-alive interval for the MQTT connection in milliseconds
     * currently configured. Setting this value to 0 will disable the keep-alive
     * function for the connection. The default keep alive interval is 30,000ms.
     *
     * @return the keep alive interval
     */
    @Override
    public int getKeepAliveInterval() {
        return super.getKeepAliveInterval();
    }

    /**
     * Sets a new value in milliseconds for the connection keep-alive interval.
     * This value must be set before {@link #connect()} is called. Setting this
     * value to 0 will disable the keep-alive function.
     *
     * @param keepAliveInterval
     *            the new keep alive interval. The default value is 30,000ms.
     */
    @Override
    public void setKeepAliveInterval(int keepAliveInterval) {
        super.setKeepAliveInterval(keepAliveInterval);
    }

    /**
     * Gets the maximum offline queue size current configured. The offline
     * queues are used for temporarily holding outgoing requests while the
     * connection is being established or retried. When the connection is
     * established, offline queue messages will be sent out as usual. They can
     * be useful for dealing with transient connection failures by allowing the
     * application to continuously send requests while the connection is being
     * established. Each type of request, namely publish, subscribe, and
     * unsubscribe, has their own offline queue. The default offline queue size
     * is 64. Setting it to 0 will disable the offline queues.
     *
     * @return the max offline queue size
     */
    @Override
    public int getMaxOfflineQueueSize() {
        return super.getMaxOfflineQueueSize();
    }

    /**
     * Sets a new value for the maximum offline queue size. This value must be
     * set before {@link #connect()} is called. Setting it to 0 will disable the
     * offline queues.
     *
     * @param maxOfflineQueueSize
     *            the new maximum offline queue size. The default value is 64.
     */
    @Override
    public void setMaxOfflineQueueSize(int maxOfflineQueueSize) {
        super.setMaxOfflineQueueSize(maxOfflineQueueSize);
    }

    /**
     * Gets the Last Will and Testament message currently configured. The Last
     * Will and Testament message with configured payload will be published when
     * the client connection is lost or terminated ungracefully, i.e. not
     * through {@link #disconnect()}.
     *
     * @return the will message
     */
    @Override
    public JoylinkMessage getWillMessage() {
        return super.getWillMessage();
    }

    /**
     * Sets whether the client and server should establish a clean session on each connection.
     * If false, the server should attempt to persist the client's state between connections.
     * This must be set before {@link #connect()} is called.
     *
     * @param cleanSession
     *            If true, the server starts a clean session with the client on each connection.
     *            If false, the server should persist the client's state between connections.
     */
    @Override
    public void setCleanSession(boolean cleanSession) { super.setCleanSession(cleanSession); }

    /**
     * Gets whether each connection is a clean session. See also {@link #setCleanSession(boolean)}.
     * @return whether each connection to the server should be a clean session.
     */
    @Override
    public boolean isCleanSession() { return super.isCleanSession(); }

    /**
     * Sets a new Last Will and Testament message. The message must be set
     * before {@link #connect()} is called. By default, Last Will and Testament
     * message is not sent.
     *
     * @param willMessage
     *            the new Last Will and Testament message message. The default
     *            value is {@code null}.
     */
    @Override
    public void setWillMessage(JoylinkMessage willMessage) {
        super.setWillMessage(willMessage);
    }

    /**
     * Connect the client to the server. This is a blocking call, so the calling
     * thread will be blocked until the operation succeeded or failed.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @throws JoylinkException
     *             exception thrown if the connection operation fails
     */
    @Override
    public void connect() throws JoylinkException {
        super.connect();
    }

    /**
     * Connect the client to the server. This is a blocking call, so the calling
     * thread will be blocked until the operation succeeded, failed, or timed
     * out.
     *
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @throws JoylinkException
     *             exception thrown if the operation fails
     * @throws JoylinkTimeoutException
     *             exception thrown if the operation times out
     */
    @Override
    public void connect(long timeout) throws JoylinkException, JoylinkTimeoutException {
        super.connect(timeout);
    }

    /**
     * Connect the client to the server. This call can be either blocking or
     * non-blocking specified by the {@code blocking} argument. For blocking
     * calls, the calling thread is blocked until the operation completed,
     * failed, or timed out; for non-blocking calls, the calling thread will not
     * be blocked while the connection is being established.
     *
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @param blocking
     *            whether the call should be blocking or non-blocking
     * @throws JoylinkException
     *             exception thrown if the operation fails
     * @throws JoylinkTimeoutException
     *             exception thrown if the operation times out
     */
    @Override
    public void connect(long timeout, boolean blocking) throws JoylinkException, JoylinkTimeoutException {
        super.connect(timeout, blocking);
    }

    /**
     * Disconnect the client from the server. This is a blocking call, so the
     * calling thread will be blocked until the operation succeeded or failed.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @throws JoylinkException
     *             exception thrown if the operation fails
     */
    @Override
    public void disconnect() throws JoylinkException {
        super.disconnect();
    }

    /**
     * Disconnect the client from the server. This is a blocking call, so the
     * calling thread will be blocked until the operation succeeded, failed, or
     * timed out.
     *
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @throws JoylinkException
     *             exception thrown if the operation fails
     * @throws JoylinkTimeoutException
     *             exception thrown if the operation times out
     */
    @Override
    public void disconnect(long timeout) throws JoylinkException, JoylinkTimeoutException {
        super.disconnect(timeout);
    }

    /**
     * Disconnect the client from the server. This call can be either blocking
     * or non-blocking specified by the {@code blocking} argument. For blocking
     * calls, the calling thread is blocked until the operation completed,
     * failed, or timed out; for non-blocking calls, the calling thread will not
     * be blocked while the connection is being terminated.
     *
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @param blocking
     *            whether the call should be blocking or non-blocking
     * @throws JoylinkException
     *             exception thrown if the operation fails
     * @throws JoylinkTimeoutException
     *             exception thrown if the operation times out
     */
    @Override
    public void disconnect(long timeout, boolean blocking) throws JoylinkException, JoylinkTimeoutException {
        super.disconnect(timeout, blocking);
    }

    /**
     * Publishes the payload to a given topic. This is a blocking call so the
     * calling thread is blocked until the publish operation succeeded or
     * failed. MQTT QoS0 is used for publishing the payload.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @param topic
     *            the topic to be published to
     * @param payload
     *            the payload to be published
     * @throws JoylinkException
     *             exception thrown if the publish operation fails
     */
    @Override
    public void publish(String topic, String payload) throws JoylinkException {
        super.publish(topic, payload);
    }

    /**
     * Publishes the payload to a given topic. This is a blocking call so the
     * calling thread is blocked until the publish operation succeeded, failed,
     * or the specified timeout has elapsed. MQTT QoS0 is used for publishing
     * the payload.
     *
     * @param topic
     *            the topic to be published to
     * @param payload
     *            the payload to be published
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails
     * @throws JoylinkTimeoutException
     *             the exception thrown if the publish operation times out
     */
    @Override
    public void publish(String topic, String payload, long timeout) throws JoylinkException, JoylinkTimeoutException {
        super.publish(topic, payload, timeout);
    }

    /**
     * Publishes the payload to a given topic. This is a blocking call so the
     * calling thread is blocked until the publish operation succeeded or
     * failed.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @param topic
     *            the topic to be published to
     * @param qos
     *            the MQTT QoS used for publishing
     * @param payload
     *            the payload to be published
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails
     */
    @Override
    public void publish(String topic, JoylinkQos qos, String payload) throws JoylinkException {
        super.publish(topic, qos, payload);
    }

    /**
     * Publishes the payload to a given topic. This is a blocking call so the
     * calling thread is blocked until the publish operation succeeded, failed,
     * or the specified timeout has elapsed.
     *
     * @param topic
     *            the topic to be published to
     * @param qos
     *            the MQTT QoS used for publishing
     * @param payload
     *            the payload to be published
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails
     * @throws JoylinkTimeoutException
     *             the exception thrown if the publish operation times out
     */
    @Override
    public void publish(String topic, JoylinkQos qos, String payload, long timeout) throws JoylinkException,
            JoylinkTimeoutException {
        super.publish(topic, qos, payload, timeout);
    }

    /**
     * Publishes the raw payload to a given topic. This is a blocking call so
     * the calling thread is blocked until the publish operation succeeded or
     * failed. MQTT QoS0 is used for publishing the payload.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @param topic
     *            the topic to be published to
     * @param payload
     *            the payload to be published
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails
     */
    @Override
    public void publish(String topic, byte[] payload) throws JoylinkException {
        super.publish(topic, payload);
    }

    /**
     * Publishes the raw payload to a given topic. This is a blocking call so
     * the calling thread is blocked until the publish operation succeeded,
     * failed, or the specified timeout has elapsed. MQTT QoS0 is used for
     * publishing the payload.
     *
     * @param topic
     *            the topic to be published to
     * @param payload
     *            the payload to be published
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails
     * @throws JoylinkTimeoutException
     *             the exception thrown if the publish operation times out
     */
    @Override
    public void publish(String topic, byte[] payload, long timeout) throws JoylinkException, JoylinkTimeoutException {
        super.publish(topic, payload, timeout);
    }

    /**
     * Publishes the raw payload to a given topic. This is a blocking call so
     * the calling thread is blocked until the publish operation is succeeded or
     * failed.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @param topic
     *            the topic to be published to
     * @param qos
     *            the MQTT QoS used for publishing
     * @param payload
     *            the payload to be published
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails
     */
    @Override
    public void publish(String topic, JoylinkQos qos, byte[] payload) throws JoylinkException {
        super.publish(topic, qos, payload);
    }

    /**
     * Publishes the raw payload to a given topic. This is a blocking call so
     * the calling thread is blocked until the publish operation is succeeded,
     * failed, or the specified timeout has elapsed.
     *
     * @param topic
     *            the topic to be published to
     * @param qos
     *            the MQTT QoS used for publishing
     * @param payload
     *            the payload to be published
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails
     * @throws JoylinkTimeoutException
     *             the exception thrown if the publish operation times out
     */
    @Override
    public void publish(String topic, JoylinkQos qos, byte[] payload, long timeout) throws JoylinkException,
            JoylinkTimeoutException {
        super.publish(topic, qos, payload, timeout);
    }

    /**
     * Publishes the payload to a given topic. Topic, MQTT QoS, and payload are
     * given in the {@code message} argument. This is a non-blocking call so it
     * immediately returns once the operation has been queued in the system. The
     * result of the operation will be notified through the callback functions,
     * namely {@link JoylinkMessage#onSuccess} and
     * {@link JoylinkMessage#onFailure}, one of which will be invoked after the
     * operation succeeded or failed respectively. The default implementation
     * for the callback functions in {@link JoylinkMessage} does nothing. The
     * user could override one or more of these functions through subclassing.
     *
     * @param message
     *            the message, including the topic, MQTT QoS, and payload, to be
     *            published
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails to be
     *             queued
     */
    @Override
    public void publish(JoylinkMessage message) throws JoylinkException {
        super.publish(message);
    }

    /**
     * Publishes the payload to a given topic. Topic, MQTT QoS, and payload are
     * given in the {@code message} argument. This is a non-blocking call so it
     * immediately returns once the operation has been queued in the system. The
     * result of the operation will be notified through the callback functions,
     * namely {@link JoylinkMessage#onSuccess}, {@link JoylinkMessage#onFailure},
     * and {@link JoylinkMessage#onTimeout}, one of which will be invoked after
     * the operation succeeded, failed, or timed out respectively. The user
     * could override one or more of these functions through subclassing.
     *
     * @param message
     *            the message, including the topic, MQTT QoS, and payload, to be
     *            published
     * @param timeout
     *            the timeout in milliseconds for the operation to be considered
     *            timed out
     * @throws JoylinkException
     *             the exception thrown if the publish operation fails to be
     *             queued
     */
    @Override
    public void publish(JoylinkMessage message, long timeout) throws JoylinkException {
        super.publish(message, timeout);
    }

    /**
     * Subscribes to a given topic. Topic and MQTT QoS are given in the
     * {@code topic} argument. This call can be either blocking or non-blocking
     * specified by the {@code blocking} argument. For blocking calls, the
     * calling thread is blocked until the subscribe operation completed or
     * failed; for non-blocking calls, the result of the operation will be
     * notified through the callback functions, namely
     * {@link JoylinkTopic#onSuccess} and {@link JoylinkTopic#onFailure}, one of
     * which will be invoked after the operation succeeded or failed
     * respectively. For both blocking and non-blocking calls, callback function
     * {@link JoylinkTopic#onMessage} is invoked when subscribed message arrives.
     * The default implementation for the callback functions in
     * {@link JoylinkTopic} does nothing. The user could override one or more of
     * these functions through subclassing.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @param topic
     *            the topic to subscribe to
     * @param blocking
     *            whether the call should be blocking or non-blocking
     * @throws JoylinkException
     *             the exception thrown if the subscribe operation fails
     *             (blocking) or fails to be queued (non-blocking)
     */
    @Override
    public void subscribe(JoylinkTopic topic, boolean blocking) throws JoylinkException {
        super.subscribe(topic, blocking);
    }

    /**
     * Subscribes to a given topic. Topic and MQTT QoS are given in the
     * {@code topic} argument. This call can be either blocking or non-blocking
     * specified by the {@code blocking} argument. For blocking call, the
     * calling thread is blocked until the subscribe operation completed,
     * failed, or timed out; for non-blocking call, the result of the operation
     * will be notified through the callback functions, namely
     * {@link JoylinkTopic#onSuccess}, {@link JoylinkTopic#onFailure} and
     * {@link JoylinkTopic#onTimeout}, one of which will be invoked after the
     * operation succeeded, failed, or timed out respectively. For both blocking
     * and non-blocking calls, callback function {@link JoylinkTopic#onMessage}
     * is invoked when subscribed message arrives. The default implementation
     * for the callback functions in {@link JoylinkTopic} does nothing. The user
     * could override one or more of these functions through subclassing.
     *
     * @param topic
     *            the topic to subscribe to
     * @param timeout
     *            the timeout in milliseconds for the operation to be considered
     *            timed out
     * @param blocking
     *            whether the call should be blocking or non-blocking
     * @throws JoylinkException
     *             the exception thrown if the subscribe operation fails
     *             (blocking) or fails to be queued (non-blocking)
     * @throws JoylinkTimeoutException
     *             the exception thrown if the subscribe operation times out.
     *             This exception is not thrown if the call is non-blocking;
     *             {@link JoylinkTopic#onTimeout} will be invoked instead if
     *             timeout happens.
     */
    @Override
    public void subscribe(JoylinkTopic topic, long timeout, boolean blocking) throws JoylinkException,
            JoylinkTimeoutException {
        super.subscribe(topic, timeout, blocking);
    }

    /**
     * Subscribes to a given topic. Topic and MQTT QoS are given in the
     * {@code topic} argument. This is a non-blocking call so it immediately
     * returns once is the operation has been queued in the system. The result
     * of the operation will be notified through the callback functions, namely
     * {@link JoylinkTopic#onSuccess} and {@link JoylinkTopic#onFailure}, one of
     * which will be invoked after the operation succeeded or failed
     * respectively. Another callback function, {@link JoylinkTopic#onMessage},
     * is invoked when subscribed message arrives. The default implementation
     * for the callback functions in {@link JoylinkTopic} does nothing. The user
     * could override one or more of these functions through sub-classing.
     *
     * @param topic
     *            the topic to subscribe to
     * @throws JoylinkException
     *             the exception thrown if the subscribe operation fails to be
     *             queued
     */
    @Override
    public void subscribe(JoylinkTopic topic) throws JoylinkException {
        super.subscribe(topic);
    }

    /**
     * Subscribes to a given topic. Topic and MQTT QoS are given in the
     * {@code topic} argument. This is a non-blocking call so it immediately
     * returns once is the operation has been queued in the system. The result
     * of the operation will be notified through the callback functions, namely
     * {@link JoylinkTopic#onSuccess}, {@link JoylinkTopic#onFailure}, and
     * {@link JoylinkTopic#onTimeout}, one of which will be invoked after the
     * operation succeeded, failed, or timed out respectively. Another callback
     * function, {@link JoylinkTopic#onMessage}, is invoked when subscribed
     * message arrives. The default implementation for the callback functions in
     * {@link JoylinkTopic} does nothing. The user could override one or more of
     * these functions through sub-classing.
     *
     * @param topic
     *            the topic to subscribe to
     * @param timeout
     *            the timeout in milliseconds for the operation to be considered
     *            timed out
     * @throws JoylinkException
     *             the exception thrown if the subscribe operation fails to be
     *             queued
     */
    @Override
    public void subscribe(JoylinkTopic topic, long timeout) throws JoylinkException {
        super.subscribe(topic, timeout);
    }

    /**
     * Unsubscribes to a given topic. This is a blocking call, so the calling
     * thread is blocked until the unsubscribe operation completed or failed.
     * <p>
     * Note: Blocking API call without specifying a timeout, in very rare cases,
     * can block the calling thread indefinitely, if the server response is not
     * received or lost. Use the alternative APIs with timeout for applications
     * that expect responses within fixed duration.
     * </p>
     *
     * @param topic
     *            the topic to unsubscribe to
     * @throws JoylinkException
     *             the exception thrown if the unsubscribe operation fails
     */
    @Override
    public void unsubscribe(String topic) throws JoylinkException {
        super.unsubscribe(topic);
    }

    /**
     * Unsubscribes to a given topic. This is a blocking call, so the calling
     * thread is blocked until the unsubscribe operation completed, failed, or
     * the specified timeout has elapsed.
     *
     * @param topic
     *            the topic to unsubscribe to
     * @param timeout
     *            the timeout in milliseconds that the calling thread will wait
     * @throws JoylinkException
     *             the exception thrown if the unsubscribe operation fails
     * @throws JoylinkTimeoutException
     *             the exception thrown if the unsubscribe operation times out
     */
    @Override
    public void unsubscribe(String topic, long timeout) throws JoylinkException, JoylinkTimeoutException {
        super.unsubscribe(topic, timeout);
    }

    /**
     * Unsubscribes to a given topic. This is a non-blocking call so it
     * immediately returns once the operation has been queued in the system. The
     * result of the operation will be notified through the callback functions,
     * namely {@link JoylinkTopic#onSuccess} and {@link JoylinkTopic#onFailure},
     * one of which will be invoked after the operation succeeded or failed
     * respectively. The default implementation for the callback functions in
     * {@link JoylinkTopic} does nothing. The user could override one or more of
     * these functions through subclassing.
     * 
     * @param topic
     *            the topic to unsubscribe to
     * @throws JoylinkException
     *             the exception thrown if the unsubscribe operation fails to be
     *             queued
     */
    @Override
    public void unsubscribe(JoylinkTopic topic) throws JoylinkException {
        super.unsubscribe(topic);
    }

    /**
     * Unsubscribes to a given topic. This is a non-blocking call so it
     * immediately returns once the operation has been queued in the system. The
     * result of the operation will be notified through the callback functions,
     * namely {@link JoylinkTopic#onSuccess}, {@link JoylinkTopic#onFailure}, and
     * {@link JoylinkTopic#onTimeout}, one of which will be invoked after the
     * operation succeeded, failed, or timed out respectively. The default
     * implementation for the callback functions in {@link JoylinkTopic} does
     * nothing. The user could override one or more of these functions through
     * subclassing.
     *
     * @param topic
     *            the topic to unsubscribe to
     * @param timeout
     *            the timeout in milliseconds for the operation to be considered
     *            timed out
     * @throws JoylinkException
     *             the exception thrown if the unsubscribe operation fails to be
     *             queued
     */
    @Override
    public void unsubscribe(JoylinkTopic topic, long timeout) throws JoylinkException {
        super.unsubscribe(topic, timeout);
    }

    /**
     * Attach a shadow device to the client. Once attached, the device, if
     * configured, will be automatically synchronized.
     *
     * @param device
     *            the device to be attached to the client
     * @throws JoylinkException
     *             the exception thrown if the attach operation fails
     */
    @Override
    public void attach(JoylinkDevice device) throws JoylinkException {
        super.attach(device);
    }

    /**
     * Attach a shadow subdevice to the client. Once attached, the device, if
     * configured, will be automatically synchronized.
     *
     * @param device
     *            the device to be attached to the client
     * @throws JoylinkException
     *             the exception thrown if the attach operation fails
     */
    @Override
    public void attach(JoylinkSubDevice device) throws JoylinkException {
        super.attach(device);
    }

    /**
     * Detach the given device from the client. Device and shadow
     * synchronization will be stopped after the device is detached from the
     * client.
     *
     * @param device
     *            the device to be detached from the client
     * @throws JoylinkException
     *             the exception thrown if the detach operation fails
     */
    @Override
    public void detach(JoylinkDevice device) throws JoylinkException {
        super.detach(device);
    }

    /**
     * Detach the given subdevice from the client. Device and shadow
     * synchronization will be stopped after the device is detached from the
     * client.
     *
     * @param device
     *            the device to be detached from the client
     * @throws JoylinkException
     *             the exception thrown if the detach operation fails
     */
    @Override
    public void detach(JoylinkSubDevice device) throws JoylinkException {
        super.detach(device);
    }

    /**
     * Gets the connection status of the connection used by the client.
     *
     * @return the connection status
     */
    @Override
    public JoylinkConnectionStatus getConnectionStatus() {
        return super.getConnectionStatus();
    }

    /**
     * This callback function is called when the connection used by the client
     * is successfully established. The user could supply a different callback
     * function via subclassing, however the default implementation should
     * always be called in the override function in order for the connection
     * retry as well as device synchronization to work properly.
     */
    @Override
    public void onConnectionSuccess() {
        super.onConnectionSuccess();
    }

    /**
     * This callback function is called when the connection used by the client
     * is temporarily lost. The user could supply a different callback function
     * via subclassing, however the default implementation should always be
     * called in the override function in order for the connection retry as well
     * as device synchronization to work properly.
     */
    @Override
    public void onConnectionFailure() {
        super.onConnectionFailure();
    }

    /**
     * This callback function is called when the connection used by the client
     * is permanently closed. The user could supply a different callback
     * function via subclassing, however the default implementation should
     * always be called in the override function in order for the connection
     * retry as well as device synchronization to work properly.
     */
    @Override
    public void onConnectionClosed() {
        super.onConnectionClosed();
    }

}
