/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.joylink5.services.iot.client.JoylinkConstant;
import com.joylink5.services.iot.client.protocol.AbstractJoylinkDevice;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.Command;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.CommandAck;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.ServiceKey;


public class JoylinkSubDevice extends AbstractJoylinkDevice {

    private static final Logger LOGGER = Logger.getLogger(JoylinkSubDevice.class.getName());

    public class SubDevInfo {
        private String name;
        private String deviceId;
        private String description;
        private String thingModelId;
        private String serviceName;
        private String hardwareId;
        private String protocolName;
        private String protocolSpec;

        public SubDevInfo(String name, String deviceId, String description, String thingModelId,
                       String serviceName, String hardwareId, String protocolName, String protocolSpec) {
            this.name = name;
            this.deviceId = deviceId;
            this.description = description;
            this.thingModelId = thingModelId;
            this.serviceName = serviceName;
            this.hardwareId = hardwareId;
            this.protocolName = protocolName;
            this.protocolSpec = protocolSpec;
        }

        public void setName(String name) {
            this.name = name;
        }

        public void setDeviceId(String deviceId) {
            this.deviceId = deviceId;
        }

        public void setDescription(String description) {
            this.description = description;
        }

        public void setThingModelId(String thingModelId) {
            this.thingModelId = thingModelId;
        }

        public void setServiceName(String serviceName) {
            this.serviceName = serviceName;
        }

        public void setHardwareId(String hardwareId) {
            this.hardwareId = hardwareId;
        }

        public void setProtocolName(String protocolName) {
            this.protocolName = protocolName;
        }

        public void setProtocolSpec(String protocolSpec) {
            this.protocolSpec = protocolSpec;
        }

        public String getInfo() {
            StringBuilder info = new StringBuilder("{");
            if (this.name != null) {
                info.append("\"").append("name").append("\":\"").append(this.name).append("\",");
            } else {
                info.append("\"").append("name").append("\":\"\",");
            }

            if (this.deviceId != null) {
                info.append("\"").append("device-id").append("\":\"").append(this.deviceId).append("\",");
            } else {
                info.append("\"").append("device-id").append("\":\"\",");
            }

            if (this.description != null) {
                info.append("\"").append("description").append("\":\"").append(this.description).append("\",");
            } else {
                info.append("\"").append("description").append("\":\"\",");
            }

            if (this.thingModelId != null) {
                info.append("\"").append("thing-model-id").append("\":\"").append(this.thingModelId).append("\",");
            } else {
                info.append("\"").append("thing-model-id").append("\":\"\",");
            }

            if (this.serviceName != null) {
                info.append("\"").append("service-name").append("\":\"").append(this.serviceName).append("\",");
            } else {
                info.append("\"").append("service-name").append("\":\"\",");
            }

            if (this.hardwareId != null) {
                info.append("\"").append("hardware-id").append("\":\"").append(this.hardwareId).append("\",");
            } else {
                info.append("\"").append("hardware-id").append("\":\"\",");
            }

            if (this.protocolName != null) {
                info.append("\"").append("protocol-name").append("\":\"").append(this.protocolName).append("\",");
            } else {
                info.append("\"").append("protocol-name").append("\":\"\",");
            }

            if (this.protocolSpec != null) { 
                info.append("\"").append("protocol-spec").append("\":\"").append(this.protocolSpec).append("\"");
            } else {
                info.append("\"").append("protocol-spec").append("\":\"\"");
            }
            info.append("}");

            return info.toString();
        }
    }
    private SubDevInfo subDevInfo;

    /**
     * Instantiates a new device instance.
     *
     * @param deviceId
     *            the device id
     */
    public JoylinkSubDevice(String deviceId) {
        super(deviceId);

        subDevInfo = new SubDevInfo(null, deviceId, null, null, null, null, null, null);
    }

    /**
     * This funtion handles setting of this subdevice information.
     *
     * @param name
     *          the name of the subdevice
     * @param deviceId
     *          the device id of the subdevice
     * @param description
     *          the description of the subdevice
     * @param thingModelId
     *          the thing model id of the subdevice
     * @param serviceName
     *          the service name of the subdevice
     * @param hardwareId
     *          the hardware id of the subdevice
     * @param protocolName
     *          the protocol name of the subdevice
     * @param protocolSpec
     *          the protocol specification of the subdevice, in form of JSON document
     *
     */
    public void setSubDevInfo(String name, String deviceId, String description, String thingModelId,
                    String serviceName, String hardwareId, String protocolName, String protocolSpec) {
        subDevInfo.setName(name);
        subDevInfo.setDeviceId(deviceId);
        subDevInfo.setDescription(description);
        subDevInfo.setThingModelId(thingModelId);
        subDevInfo.setServiceName(serviceName);
        subDevInfo.setHardwareId(hardwareId);
        subDevInfo.setProtocolName(protocolName);
        subDevInfo.setProtocolSpec(protocolSpec);
    }

    /**
     * Handles device activate, this function will be called when client connect
     * successfully, device atech.
     *
     * @return the report interval in milliseconds.
     */
    @Override
    public void activate() throws JoylinkException {
        super.activate();
        LOGGER.info("Enter Device Activate");
    }

    /**
     * Handles device deactivate, this function will be called when client connect
     * failed, connection lost and device detach.
     *
     * @return the report interval in milliseconds.
     */
    @Override
    public void deactivate() throws JoylinkException {
        super.deactivate();
        LOGGER.info("Enter Device Deactivate");
    }

    /**
     * Checks if version is enabled for device properties.
     *
     * @return true, if versioning is enabled.
     */
    @Override
    public boolean isEnableVersioning() {
        return super.isEnableVersioning();
    }

    /**
     * Sets the device properties' version to be enabled or disabled. This value
     * must be set before the device is attached to a client via the
     * {@link JoylinkMqttClient#attach(JoylinkDevice)} call.
     *
     * @param enableVersioning
     *          true to enable device properties' version; false to disable.
     */
    @Override
    public void setEnableVersioning(boolean enableVersioning) {
        super.setEnableVersioning(enableVersioning);
    }

    /**
     * This function handles collecting properties set for poroperties post.
     * The default implementation gets properties that are defined by 
     * Annotation class {JoylinkDeviceProperty}.
     *
     * One could override this default implementation.
     *
     * @return the JSON document containing property
     *         like this: {"window.windowOpen":false,"window.roomTemperature":21.462763}
     */
    @Override
    public String onPropertiesPost() {
        return super.onPropertiesPost();
    }

    /**
     * This function handles sending message of properties post.
     *
     * @param reportVersion
     *          current propeerties' version
     *
     * @param jsonProp
     *          the JSON document containing property
     *          like this: {"window.windowOpen":false,"window.roomTemperature":21.462763}
     * 
     */
    @Override
    public void sendPropertiesPost(long reportVersion, String jsonProp) {
        super.sendPropertiesPost(ServiceKey.CONN_AGENT, reportVersion, jsonProp);
    }

    /**
     * This function handles device operation of properties set.
     * The default implementation is to find and set properties
     * that are defined by Annotation class {JoylinkDeviceProperty}.
     *
     * One could override this default implementation.
     *
     * @param jsonProp
     *          the JSON document containing property and value that are needed to be set
     *          like this: {"window.windowOpen":false}
     * @return the JSON document containing property after operation 'properties set'
     *          like this: {"window.windowOpen":false}
     */
    @Override
    public String onPropertiesSet(String jsonProp) {
        return super.onPropertiesSet(jsonProp);
    }

    /**
     * This function handles sending message of properties set response.
     *
     * @param messageId
     *          the messageId of corresponding request message
     * @param jsonProp
     *          the JSON document cotaining property and value after operation 'propeties set'
     * @param errCode
     *          error code, definition refers to {JoylinkDeviceErrorCode}
     */
    @Override
    public void sendPropertiesSetResponse(String messageId, String jsonProp, long errCode) {
        super.sendPropertiesSetResponse(ServiceKey.CONN_AGENT, messageId, jsonProp, errCode);
    }

    /**
     * This function handles device operation of properties get.
     * The default implementation is to find and get properties
     * that are defined by Annotation class {JoylinkDeviceProperty}.
     *
     * One could override this default implementation.
     *
     * @param jsonProp
     *          the JSON document containing property array
     *          like this: ["window.windowOpen"]
     * @return the JSON document containing prooperty after operation 'properties get'
     *          like this: {"window.windowOpen":false}
     */
    @Override
    public String onPropertiesGet(String jsonProp) {
        return super.onPropertiesGet(jsonProp);
    }

    /**
     * This function handles sending message of properties get response.
     *
     * @param messageId
     *          the messageId of corresponding request message
     * @param jsonProp
     *          the JSON document cotaining property and value after operation 'propeties get'
     * @param errCode
     *          error code, definition refers to {JoylinkDeviceErrorCode}
     */
    @Override
    public void sendPropertiesGetResponse(String messageId, String jsonProp, long errCode) {
        super.sendPropertiesGetResponse(ServiceKey.CONN_AGENT, messageId, jsonProp, errCode);
    }

    /**
     * This function handles device operation of functions call.
     * The default implementation is to find and get functions
     * that are defined by Annotation class {JoylinkDeviceFunction}.
     *
     * Do not include functions of OTA service.
     *
     * One could override this default implementation.
     *
     * @param key
     *          key of the function call
     * @param inParam
     *          the input paraments of function call
     *          like this: {"inparam":"hello_world"}
     * @return the JSON document containing out parameters
     *          like this: {"message":"success"}
     */
    @Override
    public String onFunctionsCall(String key, String inParam) {
        return super.onFunctionsCall(key, inParam);
    }

    /**
     * This function handles sending message of functions call response.
     *
     * @param messageId
     *          the messageId of corresponding request message
     * @param funcKey
     *          the key of corresponding request function call
     * @param outJson
     *          the JSON document cotaining out parametes
     * @param errCode
     *          error code, definition refers to {JoylinkDeviceErrorCode} 
     */
    @Override
    public void sendFunctionCallResponse(String messageId, String funcKey, String outJson, long errCode) {
        super.sendFunctionCallResponse(ServiceKey.CONN_AGENT, messageId, funcKey, outJson, errCode);
    }

    /**
     * This function handles sending message of device event post.
     *
     * @param eventKey
     *          the key of the event
     * @param paraJson
     *          the JSON document containing parameters of this event
     *
     */
    public void sendEventPost(String eventKey, String paraJson) {
        super.sendEventPost(ServiceKey.CONN_AGENT, eventKey, paraJson);
    }

    /**
     * This function handles sending message of device online state.
     *
     * @param isOnline
     *          device online state
     *
     */
    public void sendDeviceOnlineState(Boolean isOnline) {
        String evtkey;
        if (isOnline) {
            evtkey = "connection_agent.online";
        } else {
            evtkey = "connection_agent.offline";
        }

        String paraJson = "{\"devices\":[" + subDevInfo.getInfo() + "]}";

        sendEventPost(ServiceKey.CONN_AGENT, super.getClient().getClientId(), evtkey, paraJson);
    }

    /**
     * This function handles sending message of add devices.
     *
     * @param isOnline
     *          device online state
     *
     */
    public void sendEventAddDevices() {
        String evtkey = "connection_agent.add-devices";
        String paraJson = "{\"devices\":[" + subDevInfo.getInfo() + "]}";

        sendEventPost(ServiceKey.CONN_AGENT, super.getClient().getClientId(), evtkey, paraJson);
    }

    /**
     * This function handles sending message of delete devices.
     *
     * @param isOnline
     *          device online state
     *
     */
    public void sendEventDeleteDevices() {
        String evtkey = "connection_agent.delete-devices";
        String paraJson = "{\"devices\":[" + subDevInfo.getInfo() + "]}";

        sendEventPost(ServiceKey.CONN_AGENT, super.getClient().getClientId(), evtkey, paraJson);
    }

}
