/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.core;

import cn.hutool.core.util.ZipUtil;
import com.joylink5.services.iot.client.JoylinkConfig;
import com.joylink5.services.iot.client.JoylinkConnectionStatus;
import com.joylink5.services.iot.client.JoylinkConstant;
import com.joylink5.services.iot.client.JoylinkDevice;
import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkQos;
import com.joylink5.services.iot.client.JoylinkSubDevice;
import com.joylink5.services.iot.client.JoylinkTimeoutException;
import com.joylink5.services.iot.client.JoylinkTopic;
import com.joylink5.services.iot.client.protocol.AbstractJoylinkDevice;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceFunctionsCallListener;
import com.joylink5.services.iot.client.protocol.JoylinkDevicePropertiesGetListener;
import com.joylink5.services.iot.client.protocol.JoylinkDevicePropertiesSetListener;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceThingModelPostResponseListener;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.Command;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.CommandAck;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.ServiceKey;
import lombok.Getter;
import lombok.Setter;

import javax.net.ssl.SSLSocketFactory;
import java.nio.charset.StandardCharsets;
import java.security.KeyStore;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

/**
 * The actual implementation of {@code JoylinkMqttClient}.
 */
@Getter
@Setter
public abstract class AbstractJoylinkClient implements JoylinkConnectionCallback {

    private static final int DEFAULT_MQTT_PORT = 8883;

    private static final Logger LOGGER = Logger.getLogger(AbstractJoylinkClient.class.getName());

    protected final String clientId;
    protected final String clientEndpoint;
    protected final boolean clientEnableMetrics;
    protected final boolean clientEnableGzip;
    protected final JoylinkConnectionType connectionType;

    protected int port = DEFAULT_MQTT_PORT;
    protected int numOfClientThreads = JoylinkConfig.NUM_OF_CLIENT_THREADS;
    protected int connectionTimeout = JoylinkConfig.CONNECTION_TIMEOUT;
    protected int serverAckTimeout = JoylinkConfig.SERVER_ACK_TIMEOUT;
    protected int keepAliveInterval = JoylinkConfig.KEEP_ALIVE_INTERVAL;
    protected int maxConnectionRetries = JoylinkConfig.MAX_CONNECTION_RETRIES;
    protected int baseRetryDelay = JoylinkConfig.CONNECTION_BASE_RETRY_DELAY;
    protected int maxRetryDelay = JoylinkConfig.CONNECTION_MAX_RETRY_DELAY;
    protected int maxOfflineQueueSize = JoylinkConfig.MAX_OFFLINE_QUEUE_SIZE;
    protected boolean cleanSession = JoylinkConfig.CLEAN_SESSION;
    protected JoylinkMessage willMessage;

    private final ConcurrentMap<String, JoylinkTopic> subscriptions = new ConcurrentHashMap<>();
    private final ConcurrentMap<String, AbstractJoylinkDevice> devices = new ConcurrentHashMap<>();
    private final JoylinkConnection connection;

    private ScheduledExecutorService executionService;

    private JoylinkDeviceTopicManager topicManager;

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, KeyStore keyStore, String keyPassword,
                                   boolean enableSdkMetrics,boolean enableGzip) {
        this.clientEndpoint = clientEndpoint;
        this.clientId = clientId;
        this.connectionType = JoylinkConnectionType.MQTT_OVER_TLS;
        this.clientEnableMetrics = enableSdkMetrics;
        this.clientEnableGzip = enableGzip;

        try {
            connection = new JoylinkTlsConnection(this, keyStore, keyPassword);
        } catch (JoylinkException e) {
            throw new JoylinkRuntimeException(e);
        }

        this.topicManager = new JoylinkDeviceTopicManager(this);
    }
    
    protected AbstractJoylinkClient(String clientEndpoint, String clientId, KeyStore keyStore, String keyPassword) {
        // Enable Metrics by default
        this(clientEndpoint, clientId, keyStore, keyPassword, true,true);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, String iotAccessKeyId,
                                   String iotSecretAccessKey, String sessionToken, boolean enableSdkMetrics,boolean enableGzip) {
        //setting the region blank to ensure it's determined from the clientEndpoint
        this(clientEndpoint, clientId, iotAccessKeyId, iotSecretAccessKey, sessionToken, "", enableSdkMetrics,enableGzip);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, String iotAccessKeyId,
                                   String iotSecretAccessKey, String sessionToken,
                                   String region, boolean enableSdkMetrics,boolean enableGzip) {
        this.clientEndpoint = clientEndpoint;
        this.clientId = clientId;
        this.connectionType = JoylinkConnectionType.MQTT_OVER_WEBSOCKET;
        this.clientEnableMetrics = enableSdkMetrics;
        this.clientEnableGzip = enableGzip;
        try {
            connection = new JoylinkWebsocketConnection(this, iotAccessKeyId, iotSecretAccessKey, sessionToken, region);
        } catch (JoylinkException e) {
            throw new JoylinkRuntimeException(e);
        }

        this.topicManager = new JoylinkDeviceTopicManager(this);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, String iotAccessKeyId,
                                   String iotSecretAccessKey, String sessionToken,boolean enableGzip) {
        // Enable Metrics by default
        this(clientEndpoint, clientId, iotAccessKeyId, iotSecretAccessKey, sessionToken, true,enableGzip);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, String iotAccessKeyId,
                                   String iotSecretAccessKey, String sessionToken, String region) {
        // Enable Metrics by default
        this(clientEndpoint, clientId, iotAccessKeyId, iotSecretAccessKey, sessionToken, region, true,true);
    }

    AbstractJoylinkClient(String clientEndpoint, String clientId, JoylinkConnection connection,
                         boolean enableSdkMetrics,boolean enableGzip) {
        this.clientEndpoint = clientEndpoint;
        this.clientId = clientId;
        this.connection = connection;
        this.connectionType = null;
        this.clientEnableMetrics = enableSdkMetrics;
        this.clientEnableGzip = enableGzip;

        this.topicManager = new JoylinkDeviceTopicManager(this);
    }

    AbstractJoylinkClient(String clientEndpoint, String clientId, JoylinkConnection connection) {
        // Enable Metrics by default
        this(clientEndpoint, clientId, connection, true,true);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, SSLSocketFactory socketFactory, boolean enableSdkMetrics,boolean enableGzip) {
        this.clientEndpoint = clientEndpoint;
        this.clientId = clientId;
        this.connectionType = null;
        this.clientEnableMetrics = enableSdkMetrics;
        this.clientEnableGzip = enableGzip;

        try {
            this.connection = new JoylinkTlsConnection(this, socketFactory);
        } catch (JoylinkException e) {
            throw new JoylinkRuntimeException(e);
        }

        this.topicManager = new JoylinkDeviceTopicManager(this);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, SSLSocketFactory socketFactory) {
        this(clientEndpoint, clientId, socketFactory, true,true);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, SSLSocketFactory socketFactory, int port, boolean enableSdkMetrics,boolean enableGzip) {
        this.clientEndpoint = clientEndpoint;
        this.clientId = clientId;
        this.connectionType = JoylinkConnectionType.MQTT_OVER_TLS;
        this.port = port;
        this.clientEnableMetrics = enableSdkMetrics;
        this.clientEnableGzip = enableGzip;

        try {
            this.connection = new JoylinkTlsConnection(this, socketFactory);
        } catch (JoylinkException e) {
            throw new JoylinkRuntimeException(e);
        }

        this.topicManager = new JoylinkDeviceTopicManager(this);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, SSLSocketFactory socketFactory, int port) {
        this(clientEndpoint, clientId, socketFactory, port, true,true);
    }

    protected AbstractJoylinkClient(String clientEndpoint, String clientId, int port,boolean enableSdkMetrics,boolean enableGzip) {
        this.clientEndpoint = clientEndpoint;
        this.clientId = clientId;
        this.connectionType = JoylinkConnectionType.MQTT_OVER_TCP;
        this.port = port;
        this.clientEnableMetrics = enableSdkMetrics;
        this.clientEnableGzip = enableGzip;

        try {
            this.connection = new JoylinkTcpConnection(this);
        } catch (JoylinkException e) {
            throw new JoylinkRuntimeException(e);
        }

        this.topicManager = new JoylinkDeviceTopicManager(this);
    }

    public String genTopic(ServiceKey serviceKey, Command command, CommandAck ack) {
        return this.topicManager.getTopic(serviceKey, command, ack);
    }
    
    public void updateCredentials(String iotAccessKeyId, String iotSecretAccessKey, String sessionToken) {
        this.connection.updateCredentials(iotAccessKeyId, iotSecretAccessKey, sessionToken);
    }

    public void connect() throws JoylinkException {
        try {
            connect(0, true);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because timeout is 0
            throw new JoylinkRuntimeException(e);
        }
    }

    public void connect(long timeout) throws JoylinkException, JoylinkTimeoutException {
        connect(timeout, true);
    }

    public void connect(long timeout, boolean blocking) throws JoylinkException, JoylinkTimeoutException {
        synchronized (this) {
            if (executionService == null) {
                executionService = Executors.newScheduledThreadPool(numOfClientThreads);
            }
        }

        JoylinkCompletion completion = new JoylinkCompletion(timeout, !blocking);
        connection.connect(completion);
        completion.get(this);
    }

    public void disconnect() throws JoylinkException {
        try {
            disconnect(0, true);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because timeout is 0
            throw new JoylinkRuntimeException(e);
        }
    }

    public void disconnect(long timeout) throws JoylinkException, JoylinkTimeoutException {
        disconnect(timeout, true);
    }

    public void disconnect(long timeout, boolean blocking) throws JoylinkException, JoylinkTimeoutException {
        JoylinkCompletion completion = new JoylinkCompletion(timeout, !blocking);
        connection.disconnect(completion);
        completion.get(this);
    }

    public void publish(String topic, String payload) throws JoylinkException {
        publish(topic, JoylinkQos.QOS0, payload);
    }

    public void publish(String topic, String payload, long timeout) throws JoylinkException, JoylinkTimeoutException {
        publish(topic, JoylinkQos.QOS0, payload, timeout);
    }

    public void publish(String topic, JoylinkQos qos, String payload) throws JoylinkException {
        try {
            publish(topic, qos, payload, 0);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because timeout is 0
            throw new JoylinkRuntimeException(e);
        }
    }

    public void publish(String topic, JoylinkQos qos, String payload, long timeout)
            throws JoylinkException, JoylinkTimeoutException {
        byte[] bytes = StandardCharsets.UTF_8.encode(payload).array();
        if(clientEnableGzip){
            bytes  = ZipUtil.gzip(bytes);
        }
        JoylinkCompletion completion = new JoylinkCompletion(topic, qos, bytes, timeout);
        connection.publish(completion);
        completion.get(this);
    }

    public void publish(String topic, byte[] payload) throws JoylinkException {
        publish(topic, JoylinkQos.QOS0, payload);
    }

    public void publish(String topic, byte[] payload, long timeout) throws JoylinkException, JoylinkTimeoutException {
        publish(topic, JoylinkQos.QOS0, payload, timeout);
    }

    public void publish(String topic, JoylinkQos qos, byte[] payload) throws JoylinkException {
        try {
            publish(topic, qos, payload, 0);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because timeout is 0
            throw new JoylinkRuntimeException(e);
        }
    }

    public void publish(String topic, JoylinkQos qos, byte[] payload, long timeout)
            throws JoylinkException, JoylinkTimeoutException {
        JoylinkCompletion completion = new JoylinkCompletion(topic, qos, payload, timeout);
        connection.publish(completion);
        completion.get(this);
    }

    public void publish(JoylinkMessage message) throws JoylinkException {
        publish(message, 0);
    }

    public void publish(JoylinkMessage message, long timeout) throws JoylinkException {
        JoylinkCompletion completion = null;
        if(clientEnableGzip){
            message.setPayload(ZipUtil.gzip(message.getPayload()));
            completion = new JoylinkCompletion(message, timeout, true);
        }else{
            completion = new JoylinkCompletion(message, timeout, true);
        }

        connection.publish(completion);
        try {
            completion.get(this);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because it's asynchronous call
            throw new JoylinkRuntimeException(e);
        }
    }

    public void subscribe(JoylinkTopic topic, boolean blocking) throws JoylinkException {
        try {
            _subscribe(topic, 0, !blocking);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because timeout is 0
            throw new JoylinkRuntimeException(e);
        }
    }

    public void subscribe(JoylinkTopic topic, long timeout, boolean blocking)
            throws JoylinkException, JoylinkTimeoutException {
        _subscribe(topic, timeout, !blocking);
    }

    public void subscribe(JoylinkTopic topic) throws JoylinkException {
        subscribe(topic, 0);
    }

    public void subscribe(JoylinkTopic topic, long timeout) throws JoylinkException {
        try {
            _subscribe(topic, timeout, true);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because it's asynchronous call
            throw new JoylinkRuntimeException(e);
        }
    }

    private void _subscribe(JoylinkTopic topic, long timeout, boolean async)
            throws JoylinkException, JoylinkTimeoutException {
        JoylinkCompletion completion = new JoylinkCompletion(topic, timeout, async);
        connection.subscribe(completion);
        completion.get(this);

        subscriptions.put(topic.getTopic(), topic);
    }

    public void unsubscribe(String topic) throws JoylinkException {
        try {
            unsubscribe(topic, 0);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because timeout is 0
            throw new JoylinkRuntimeException(e);
        }
    }

    public void unsubscribe(String topic, long timeout) throws JoylinkException, JoylinkTimeoutException {
        if (subscriptions.remove(topic) == null) {
            return;
        }

        JoylinkCompletion completion = new JoylinkCompletion(topic, JoylinkQos.QOS0, timeout);
        connection.unsubscribe(completion);
        completion.get(this);
    }

    public void unsubscribe(JoylinkTopic topic) throws JoylinkException {
        unsubscribe(topic, 0);
    }

    public void unsubscribe(JoylinkTopic topic, long timeout) throws JoylinkException {
        if (subscriptions.remove(topic.getTopic()) == null) {
            return;
        }

        JoylinkCompletion completion = new JoylinkCompletion(topic, timeout, true);
        connection.unsubscribe(completion);
        try {
            completion.get(this);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because it's asynchronous call
            throw new JoylinkRuntimeException(e);
        }
    }

    public boolean topicFilterMatch(String topicFilter, String topic) {
        if (topicFilter == null || topic == null) {
            return false;
        }

        String[] filterTokens = topicFilter.split("/");
        String[] topicTokens = topic.split("/");
        if (filterTokens.length > topicTokens.length) {
            return false;
        }

        for (int i = 0; i < filterTokens.length; i++) {
            if (filterTokens[i].equals("#")) {
                // '#' must be the last character
                return ((i + 1) == filterTokens.length);
            }

            if (!(filterTokens[i].equals(topicTokens[i]) || filterTokens[i].equals("+"))) {
                return false;
            }
        }

        return (filterTokens.length == topicTokens.length);
    }

    public ServiceKey topicGetServiceKey(String topic) {
        String[] topicTokens = topic.split("/");
        if (topicTokens[4].equals(JoylinkConstant.TOPIC_SER_KEY_AGENT)) {
            return ServiceKey.CONN_AGENT;
        } else if (topicTokens[4].equals(JoylinkConstant.TOPIC_SER_KEY_OTA)) {
            return ServiceKey.OTA;
        } else {
            return null;
        }
    }

    public void dispatch(final JoylinkMessage message) {
        boolean matches = false;

        for (String topicFilter : subscriptions.keySet()) {
            if (topicFilterMatch(topicFilter, message.getTopic())) {
                final JoylinkTopic topic = subscriptions.get(topicFilter);
                scheduleTask(new Runnable() {
                    @Override
                    public void run() {
                        topic.onMessage(message);
                    }
                });
                matches = true;
            }
        }

        if (!matches) {
            LOGGER.warning("Unexpected message received from topic " + message.getTopic());
        }
    }

    public void attach(JoylinkDevice device) throws JoylinkException {
        if (devices.putIfAbsent(device.getDeviceId(), device) != null) {
            return;
        }

        device.setClient(this);

        // start the shadow sync task if the connection is already established
        if (getConnectionStatus().equals(JoylinkConnectionStatus.CONNECTED)) {
            device.activate();
        }
    }

    public void attach(JoylinkSubDevice device) throws JoylinkException {
        if (devices.putIfAbsent(device.getDeviceId(), device) != null) {
            return;
        }

        device.setClient(this);

        // start the shadow sync task if the connection is already established
        if (getConnectionStatus().equals(JoylinkConnectionStatus.CONNECTED)) {
            device.activate();
        }
    }

    public void detach(JoylinkDevice device) throws JoylinkException {
        if (devices.remove(device.getDeviceId()) == null) {
            return;
        }

        device.deactivate();
    }

    public void detach(JoylinkSubDevice device) throws JoylinkException {
        if (devices.remove(device.getDeviceId()) == null) {
            return;
        }

        device.deactivate();
    }

    public JoylinkConnectionStatus getConnectionStatus() {
        if (connection != null) {
            return connection.getConnectionStatus();
        } else {
            return JoylinkConnectionStatus.DISCONNECTED;
        }
    }

    @Override
    public void onConnectionSuccess() {
        LOGGER.info("Client connection active: " + clientId);

        try {
            // resubscribe all the subscriptions
            /*
            for (JoylinkTopic topic : subscriptions.values()) {
                subscribe(topic, serverAckTimeout);
            }*/
            // Subscribe Topics Which Are Needed
            JoylinkTopic joylinkTopic;
            String subTopic;

            // Subscribe Thing-model Post Response
            subTopic = topicManager.getTopic(null, Command.THMD_POST, CommandAck.RESPONSE);
            joylinkTopic = new JoylinkDeviceThingModelPostResponseListener(subTopic, JoylinkQos.QOS0, this);
            subscribe(joylinkTopic, getServerAckTimeout());
            LOGGER.info("Joylink Clinet Subscribe Topic: " + joylinkTopic.getTopic() + "\r\n");

            // Subscribe Properties Get
            subTopic = topicManager.getTopic(null, Command.PROP_GET, null);
            joylinkTopic = new JoylinkDevicePropertiesGetListener(subTopic, JoylinkQos.QOS0, this);
            subscribe(joylinkTopic, getServerAckTimeout());
            LOGGER.info("Joylink Clinet Subscribe Topic: " + joylinkTopic.getTopic() + "\r\n");

            subTopic = topicManager.getTopic(ServiceKey.WILDCARD, Command.PROP_GET, null);
            joylinkTopic = new JoylinkDevicePropertiesGetListener(subTopic, JoylinkQos.QOS0, this);
            subscribe(joylinkTopic, getServerAckTimeout());
            LOGGER.info("Joylink Clinet Subscribe Topic: " + joylinkTopic.getTopic() + "\r\n");

            // Subscribe Properties Set
            subTopic = topicManager.getTopic(null, Command.PROP_SET, null);
            joylinkTopic = new JoylinkDevicePropertiesSetListener(subTopic, JoylinkQos.QOS0, this);
            subscribe(joylinkTopic, getServerAckTimeout());
            LOGGER.info("Joylink Clinet Subscribe Topic: " + joylinkTopic.getTopic() + "\r\n");

            subTopic = topicManager.getTopic(ServiceKey.WILDCARD, Command.PROP_SET, null);
            joylinkTopic = new JoylinkDevicePropertiesSetListener(subTopic, JoylinkQos.QOS0, this);
            subscribe(joylinkTopic, getServerAckTimeout());
            LOGGER.info("Joylink Clinet Subscribe Topic: " + joylinkTopic.getTopic() + "\r\n");

            // Subscribe Functions Call
            subTopic = topicManager.getTopic(null, Command.FUNC_CALL, null);
            joylinkTopic = new JoylinkDeviceFunctionsCallListener(subTopic, JoylinkQos.QOS0, this);
            subscribe(joylinkTopic, getServerAckTimeout());
            LOGGER.info("Joylink Clinet Subscribe Topic: " + joylinkTopic.getTopic() + "\r\n");

            subTopic = topicManager.getTopic(ServiceKey.WILDCARD, Command.FUNC_CALL, null);
            joylinkTopic = new JoylinkDeviceFunctionsCallListener(subTopic, JoylinkQos.QOS0, this);
            subscribe(joylinkTopic, getServerAckTimeout());
            LOGGER.info("Joylink Clinet Subscribe Topic: " + joylinkTopic.getTopic() + "\r\n");

            // start device sync
            for (AbstractJoylinkDevice device : devices.values()) {
                device.activate();
            }
        } catch (JoylinkException e) {
            // connection couldn't be fully recovered, disconnecting
            LOGGER.warning("Failed to complete subscriptions while client is active, will disconnect");
            try {
                connection.disconnect(null);
            } catch (JoylinkException de) {
                // ignore disconnect errors
            }
        }
    }

    @Override
    public void onConnectionFailure() {
        LOGGER.info("Client connection lost: " + clientId);

        // stop device sync
        for (AbstractJoylinkDevice device : devices.values()) {
            try {
                device.deactivate();
            } catch (JoylinkException e) {
                // ignore errors from deactivate() as the connection is lost
                LOGGER.warning("Failed to deactive all the devices, ignoring the error");
            }
        }
    }

    @Override
    public void onConnectionClosed() {
        LOGGER.info("Client connection closed: " + clientId);

        // stop device sync
        for (AbstractJoylinkDevice device : devices.values()) {
            try {
                device.deactivate();
            } catch (JoylinkException e) {
                // ignore errors from deactivate() as the connection is lost
                LOGGER.warning("Failed to deactive all the devices, ignoring the error");
            }
        }

        subscriptions.clear();
        devices.clear();

        executionService.shutdown();
    }

    public Future<?> scheduleTask(Runnable runnable) {
        return scheduleTimeoutTask(runnable, 0);
    }

    public Future<?> scheduleTimeoutTask(Runnable runnable, long timeout) {
        if (executionService == null) {
            throw new JoylinkRuntimeException("Client is not connected");
        }
        return executionService.schedule(runnable, timeout, TimeUnit.MILLISECONDS);
    }

    public Future<?> scheduleRoutineTask(Runnable runnable, long initialDelay, long period) {
        if (executionService == null) {
            throw new JoylinkRuntimeException("Client is not connected");
        }
        return executionService.scheduleAtFixedRate(runnable, initialDelay, period, TimeUnit.MILLISECONDS);
    }

}
