/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.core;

import java.util.HashSet;
import java.util.Set;

import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.mqtt.JoylinkMqttConnection;
import com.joylink5.services.iot.client.util.JoylinkWebSocketUrlSigner;

/**
 * This is a thin layer on top of {@link JoylinkMqttConnection} that provides a
 * WebSocket based communication channel to the MQTT implementation.
 */
public class JoylinkWebsocketConnection extends JoylinkMqttConnection {

    private JoylinkWebSocketUrlSigner urlSigner;

    public JoylinkWebsocketConnection(AbstractJoylinkClient client, String iotAccessKeyId, String iotSecretAccessKey)
            throws JoylinkException {
        this(client, iotAccessKeyId, iotSecretAccessKey, null);
    }

    public JoylinkWebsocketConnection(AbstractJoylinkClient client, String iotAccessKeyId, String iotSecretAccessKey,
                                     String sessionToken, String region) throws JoylinkException {
        super(client, null, "wss://" + client.getClientEndpoint() + ":443");

        // Port number must be included in the endpoint for signing otherwise
        // the signature verification will fail. This is because the Paho client
        // library always includes port number in the host line of the
        // HTTP request header.
        urlSigner = new JoylinkWebSocketUrlSigner(client.getClientEndpoint() + ":443", region);
        urlSigner.updateCredentials(iotAccessKeyId, iotSecretAccessKey, sessionToken);
    }

    public JoylinkWebsocketConnection(AbstractJoylinkClient client, String iotAccessKeyId, String iotSecretAccessKey,
            String sessionToken) throws JoylinkException {
        //setting the region blank to ensure it's determined from the client Endpoint
        this(client, iotAccessKeyId, iotSecretAccessKey, sessionToken, "");
    }

    @Override
    public void updateCredentials(String iotAccessKeyId, String iotSecretAccessKey, String sessionToken) {
        urlSigner.updateCredentials(iotAccessKeyId, iotSecretAccessKey, sessionToken);
    }

    @Override
    public Set<String> getServerUris() {
        Set<String> uris = new HashSet<>();
        try {
            uris.add(urlSigner.getSignedUrl(null));
        } catch (JoylinkException e) {
            throw new JoylinkRuntimeException(e);
        }

        return uris;
    }

}
