/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.mqtt;

import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.core.AbstractJoylinkClient;
import com.joylink5.services.iot.client.core.JoylinkConnection;
import com.joylink5.services.iot.client.core.JoylinkMessageCallback;
import com.joylink5.services.iot.client.core.JoylinkRetryableException;
import lombok.Getter;
import lombok.Setter;
import org.eclipse.paho.client.mqttv3.MqttAsyncClient;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;

import javax.net.SocketFactory;
import java.util.HashSet;
import java.util.Set;

/**
 * This class extends {@link JoylinkConnection} to provide the basic MQTT pub/sub
 * functionalities using the Paho MQTT library.
 */
@Getter
@Setter
public class JoylinkMqttConnection extends JoylinkConnection {

    private final SocketFactory socketFactory;

    private MqttAsyncClient mqttClient;
    private JoylinkMqttMessageListener messageListener;
    private JoylinkMqttClientListener clientListener;

    public JoylinkMqttConnection(AbstractJoylinkClient client, SocketFactory socketFactory, String serverUri)
            throws JoylinkException {
        super(client);

        this.socketFactory = socketFactory;

        messageListener = new JoylinkMqttMessageListener(client);
        clientListener = new JoylinkMqttClientListener(client);

        try {
            mqttClient = new MqttAsyncClient(serverUri, client.getClientId(), new MemoryPersistence());
            mqttClient.setCallback(clientListener);
        } catch (MqttException e) {
            throw new JoylinkException(e);
        }
    }

    JoylinkMqttConnection(AbstractJoylinkClient client, MqttAsyncClient mqttClient) throws JoylinkException {
        super(client);
        this.mqttClient = mqttClient;
        this.socketFactory = null;
    }

    public JoylinkMqttConnection(AbstractJoylinkClient client, String serverUri)
            throws JoylinkException {
        super(client);

        this.socketFactory = null;

        messageListener = new JoylinkMqttMessageListener(client);
        clientListener = new JoylinkMqttClientListener(client);

        try {
            mqttClient = new MqttAsyncClient(serverUri, client.getClientId(), new MemoryPersistence());
            mqttClient.setCallback(clientListener);
        } catch (MqttException e) {
            throw new JoylinkException(e);
        }
    }

    public void openConnection(JoylinkMessageCallback callback) throws JoylinkException {
        try {
            JoylinkMqttConnectionListener connectionListener = new JoylinkMqttConnectionListener(client, true, callback);
            MqttConnectOptions options = buildMqttConnectOptions(client, socketFactory);
            mqttClient.connect(options, null, connectionListener);
        } catch (MqttException e) {
            throw new JoylinkException(e);
        }
    }

    public void closeConnection(JoylinkMessageCallback callback) throws JoylinkException {
        try {
            JoylinkMqttConnectionListener connectionListener = new JoylinkMqttConnectionListener(client, false, callback);
            mqttClient.disconnect(0, null, connectionListener);
        } catch (MqttException e) {
            throw new JoylinkException(e);
        }
    }

    @Override
    public void publishMessage(JoylinkMessage message) throws JoylinkException, JoylinkRetryableException {
        String topic = message.getTopic();
        MqttMessage mqttMessage = new MqttMessage(message.getPayload());
        mqttMessage.setQos(message.getQos().getValue());

        try {
            mqttClient.publish(topic, mqttMessage, message, messageListener);
        } catch (MqttException e) {
            if (e.getReasonCode() == MqttException.REASON_CODE_CLIENT_NOT_CONNECTED) {
                throw new JoylinkRetryableException(e);
            } else {
                throw new JoylinkException(e);
            }
        }
    }

    @Override
    public void subscribeTopic(JoylinkMessage message) throws JoylinkException, JoylinkRetryableException {
        try {
            mqttClient.subscribe(message.getTopic(), message.getQos().getValue(), message, messageListener);
        } catch (MqttException e) {
            if (e.getReasonCode() == MqttException.REASON_CODE_CLIENT_NOT_CONNECTED) {
                throw new JoylinkRetryableException(e);
            } else {
                throw new JoylinkException(e);
            }
        }
    }

    @Override
    public void unsubscribeTopic(JoylinkMessage message) throws JoylinkException, JoylinkRetryableException {
        try {
            mqttClient.unsubscribe(message.getTopic(), message, messageListener);
        } catch (MqttException e) {
            if (e.getReasonCode() == MqttException.REASON_CODE_CLIENT_NOT_CONNECTED) {
                throw new JoylinkRetryableException(e);
            } else {
                throw new JoylinkException(e);
            }
        }
    }

    public Set<String> getServerUris() {
        return new HashSet<>();
    }

    private MqttConnectOptions buildMqttConnectOptions(AbstractJoylinkClient client, SocketFactory socketFactory) {
        MqttConnectOptions options = new MqttConnectOptions();

        if (socketFactory != null) {
            options.setSocketFactory(socketFactory);
        }
        options.setCleanSession(client.isCleanSession());
        options.setConnectionTimeout(client.getConnectionTimeout() / 1000);
        options.setKeepAliveInterval(client.getKeepAliveInterval() / 1000);
        options.setUserName(client.getClientId());

        Set<String> serverUris = getServerUris();
        if (serverUris != null && !serverUris.isEmpty()) {
            String[] uriArray = new String[serverUris.size()];
            serverUris.toArray(uriArray);
            options.setServerURIs(uriArray);
        }

        if (client.getWillMessage() != null) {
            JoylinkMessage message = client.getWillMessage();

            options.setWill(message.getTopic(), message.getPayload(), message.getQos().getValue(), false);
        }

        return options;
    }

}
