/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.protocol;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.Iterator;


import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.nio.channels.NonWritableChannelException;

import com.joylink5.services.iot.client.JoylinkConfig;
import com.joylink5.services.iot.client.JoylinkDevice;
import com.joylink5.services.iot.client.JoylinkDeviceProperty;
import com.joylink5.services.iot.client.JoylinkDeviceFunction;
import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkQos;
import com.joylink5.services.iot.client.JoylinkTimeoutException;
import com.joylink5.services.iot.client.JoylinkTopic;
import com.joylink5.services.iot.client.JoylinkConstant;
import com.joylink5.services.iot.client.JoylinkDeviceErrorCode;
import com.joylink5.services.iot.client.JoylinkOtaFunctions;
import com.joylink5.services.iot.client.JoylinkConnectionAgentFunctions;
import com.joylink5.services.iot.client.core.AbstractJoylinkClient;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.Command;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.CommandAck;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.ServiceKey;

import org.bouncycastle.jcajce.provider.digest.Tiger;

import com.fasterxml.jackson.core.JsonProcessingException;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.deser.impl.CreatorCandidate.Param;
import com.fasterxml.jackson.databind.module.SimpleModule;
import com.fasterxml.jackson.databind.ser.std.StdArraySerializers;

import lombok.Getter;
import lombok.Setter;
import lombok.extern.java.Log;

/**
 * The actual implementation of {@link JoylinkDevice}.
 */
@Getter
@Setter
public abstract class AbstractJoylinkDevice {

    private static final Logger LOGGER = Logger.getLogger(AbstractJoylinkDevice.class.getName());

    protected final String deviceId;

    protected boolean enableVersioning = JoylinkConfig.DEVICE_ENABLE_VERSIONING;

    private final Map<String, Field> properties;
    private final Map<String, Field> functions;
    private final JoylinkDeviceCommandManager commandManager;

    private AbstractJoylinkClient client;
    private AtomicLong localVersion;

    private JoylinkDeviceErrorCode errorCode;

    // OTA Properties
    private String otaPropModel;
    private Boolean otaPropAcceptDiff = false;
    private String otaPropSigAlg;
    private List<Map<String, String>> otaPropList;

    // ota functions listener
    private JoylinkOtaFunctions otaFunctionsListener;

    // connection_agent functions listener
    private JoylinkConnectionAgentFunctions connAgentFunctionsListener;

    protected AbstractJoylinkDevice(String deviceId) {
        this.deviceId = deviceId;

        properties = getDeviceProperties();
        functions = getDeviceFunctions();
        commandManager = new JoylinkDeviceCommandManager(this);

        localVersion = new AtomicLong(0);

        errorCode = JoylinkDeviceErrorCode.SUCESS;

        this.otaPropList = new ArrayList<>();
    }

    protected AbstractJoylinkDevice getDevice() {
        return this;
    }

/**********************************************************************************************
 * Property Operation
 **********************************************************************************************/
    protected String onPropertiesSet(String jsonProp) {
        // synchronized block to serialize device accesses
        synchronized (this) {
            String propJson = null;
            ObjectMapper mapper = new ObjectMapper();
            Map<String, Object> propMap = new HashMap<>();
            try {
                JsonNode node = mapper.readTree(jsonProp);
                if (node == null) {
                    errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                    LOGGER.warning("Invalid Property Set jsonProp:" + jsonProp);
                    return null;
                }
                for (Iterator<String> it = node.fieldNames(); it.hasNext();) {
                    String property = it.next();
                    Field field = this.properties.get(property);
                    JsonNode fieldNode = node.get(property);
                    if (field == null || fieldNode == null) {
                        LOGGER.warning("Can Not Find Property:" + property);
                        errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                        return propJson;
                    }

                    Object value = mapper.treeToValue(fieldNode, field.getType());
                    invokeSetterMethod(this, field.getName(), field.getType(), value);

                    // Get Setted Value
                    value = invokeGetterMethod(this, field);
                    propMap.put(property, value);
                }
                propJson = mapper.writeValueAsString(propMap);
            } catch (IOException e) {
                LOGGER.warning("Failed to Set Properties");
                errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return null;
            }

            return propJson;
        }
    }

    protected String onPropertiesGet(String jsonProp) {
        // synchronized block to serialize device accesses
        synchronized (this) {
            LOGGER.warning("Property Get Json:" + jsonProp);
            String propJson = null;
            ObjectMapper mapper = new ObjectMapper();
            Map<String, Object> propMap = new HashMap<>();
            try {
                JsonNode node = mapper.readTree(jsonProp);
                if (node == null) {
                    errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                    LOGGER.warning("Invalid Property Get jsonProp:" + jsonProp);
                    return null;
                }
                if (!node.isArray()) {
                    LOGGER.warning("Properties Get Json Node Type Should Be Array!");
                    errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                    return null;
                }
                for (JsonNode item : node) {
                    String property = item.asText();
                    Field field = this.properties.get(property);
                    if (field == null) {
                        LOGGER.severe("Can not find property:" + property);
                        errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                        return null;
                    }
                    Object value = invokeGetterMethod(this, field);
                    propMap.put(property, value);
                }
                propJson = mapper.writeValueAsString(propMap);
            } catch (IOException e) {
                LOGGER.warning("Failed to Get Properties");
                errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return null;
            }

            return propJson;
        }
    }

    protected String onOTAPropertiesGet(String jsonProp) {
        // synchronized block to serialize device accesses
        synchronized (this) {
            LOGGER.warning("OTA Property Get Json:" + jsonProp);
            String propJson = null;
            ObjectMapper mapper = new ObjectMapper();
            Map<String, Object> propMap = new HashMap<>();
            try {
                JsonNode node = mapper.readTree(jsonProp);
                if (node == null) {
                    errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                    LOGGER.warning("Invalid Property Get jsonProp:" + jsonProp);
                    return null;
                }
                if (!node.isArray()) {
                    LOGGER.warning("Properties Get Json Node Type Should Be Array!");
                    errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                    return null;
                }
                for (JsonNode item : node) {
                    String property = item.asText();
                    Field field = this.properties.get(property);
                    if (property.equals("ota.model")) {
                        propMap.put(property, this.otaPropModel);
                    } else if (property.equals("ota.accept-diff")) {
                        propMap.put(property, this.otaPropAcceptDiff);
                    } else if (property.equals("ota.signature-algorithm")) {
                        propMap.put(property, this.otaPropSigAlg);
                    } else if (property.equals("ota.ota-list")) {
                        propMap.put(property, this.otaPropList);
                    } else {}
                }
                propJson = mapper.writeValueAsString(propMap);
            } catch (IOException e) {
                LOGGER.warning("Failed to Get Properties");
                errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return null;
            }

            return propJson;
        }
    }

    protected String onPropertiesPost() {
        // synchronized block to serialize device accesses
        synchronized (this) {
            String propJson;
            Map<String, Object> propMap = new HashMap<>();
            try {
                for (String property : this.properties.keySet()) {
                    if (property.substring(0, 3).equals("ota")) {
                        continue;
                    }

                    Field field = this.properties.get(property);
                    Object value = invokeGetterMethod(this, field);
                    propMap.put(property, value);
                }
                if (propMap.isEmpty()) {
                    propJson = null;
                } else {
                    ObjectMapper mapper = new ObjectMapper();
                    propJson = mapper.writeValueAsString(propMap);
                }
            } catch (IOException e) {
                LOGGER.warning("[Error] Build Properties Json Failed!");
                return null;
            }

            return propJson;
        }
    }

    protected String onOtaPropertiesPost() {
        synchronized (this) {
            String propJson;
            Map<String, Object> propMap = new HashMap<>();
            try {
                propMap.put("ota.model", this.otaPropModel);
                propMap.put("ota.accept-diff", this.otaPropAcceptDiff);
                propMap.put("ota.signature-algorithm", this.otaPropSigAlg);
                propMap.put("ota.ota-list", this.otaPropList);

                ObjectMapper mapper = new ObjectMapper();
                propJson = mapper.writeValueAsString(propMap);

            } catch (IOException e) {
                LOGGER.warning("[Error] Build Properties Json Failed!");
                return null;
            }

            return propJson;
        }
    }

/**********************************************************************************************
 * Functions Call operation
 **********************************************************************************************/
    protected String onFunctionsCall(String key, String inParam) {
        // synchronized block to serialize device accesses
        synchronized (this) {
            String outParam = "";
            Field field = this.functions.get(key);
            if (field == null) {
                LOGGER.warning("Can Not Find Function:" + key);
                errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return outParam;
            }

            // Invoke Set Method to Execute Function Call
            try {
                invokeSetterMethod(this, field.getName(), String.class, inParam);
            } catch (IOException e) {
                LOGGER.warning("Call Function " + field.getName() + "Failed!");
                errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return outParam;
            }

            // Invoke Get Method to Get Function Call Out Parameters
            try {
                Object value = invokeGetterMethod(this, field);
                LOGGER.warning("Get Function Call Out Parameters:" + value.toString());
                outParam = value.toString();
            } catch (IOException e) {
                LOGGER.warning("Get Function Call " + field.getName() + " Out Parametes Failed!");
            }

            return outParam;
        }
    }

    private String getOtaInParamTaskId(String inJson) {
        ObjectMapper mapper = new ObjectMapper();
        JsonNode rootNode;
        try {
            rootNode = mapper.readTree(inJson);
            if (rootNode == null) {
                LOGGER.severe("Parse inJson failed");
                return null;
            }
            return rootNode.get("task-id").asText();
        } catch (IOException e) {
            LOGGER.severe("Encounter IOException when get task-id");
            return null;
        }
    }

    protected String onOtaFunctionsCall(String key, String inParam) {
        // synchronized block to serialize device accesses
        synchronized (this) {
            if (this.otaFunctionsListener == null) {
                errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return "";
            }
            String outParam = "";
            if (key.equals("ota.ota")) {
                this.otaFunctionsListener.onFuncOta(inParam);
            } else if (key.equals("ota.cancel-task")) {
                String taskId = getOtaInParamTaskId(inParam);
                if ((taskId == null) || (taskId.length() == 0)) {
                    LOGGER.severe("task-id is null");
                } else {
                    this.otaFunctionsListener.onFuncCancelTask(taskId);
                }
            } else if (key.equals("ota.query-ota-state")) {
                String taskId = getOtaInParamTaskId(inParam);
                if ((taskId == null) || (taskId.length() == 0)) {
                    LOGGER.severe("task-id is null");
                } else {
                    outParam = this.otaFunctionsListener.onFuncQueryState(taskId);
                }
            } else if (key.equals("ota.exec-task")) {
                String taskId = getOtaInParamTaskId(inParam);
                if ((taskId == null) || (taskId.length() == 0)) {
                    LOGGER.severe("task-id is null");
                } else {
                    this.otaFunctionsListener.onFuncExecTask(taskId);
                }
            } else {}

            return outParam;
        }
    }

    protected String onConnAgentFunctionsCall(String key, String inParam) {
        // synchronized block to serialize device accesses
        synchronized (this) {
            if (this.connAgentFunctionsListener == null) {
                errorCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return "";
            }
            String outParam = "";
            if (key.equals("connection_agent.add-devices")) {
                outParam = this.connAgentFunctionsListener.onFuncAddDevices(inParam);
            } else if (key.equals("connection_agent.delete-devices")) {
                outParam = this.connAgentFunctionsListener.onFuncDeleteDevices(inParam);
            } else if (key.equals("connection_agent.update-devices")) {
                outParam = this.connAgentFunctionsListener.onFuncUpdateDevices(inParam);
            } else if (key.equals("connection_agent.get-devices")) {
                outParam = this.connAgentFunctionsListener.onFuncGetDevices();
            } else {}

            // Build out
            String msg;
            if (errorCode.equals(JoylinkDeviceErrorCode.SUCESS)) {
                msg = "success";
            } else {
                msg = "faild";
            }
            if ((outParam == null) || (outParam.length() == 0)) {
                outParam = onBuildConnAgentFuncCallOutJson(errorCode.getValue(), msg, null);
            } else {
                outParam = onBuildConnAgentFuncCallOutJson(errorCode.getValue(), msg, outParam);
            }

            return outParam;
        }
    }

    private String onBuildConnAgentFuncCallOutJson(long code, String message, String devices) {
        StringBuilder payload = new StringBuilder("{");

        payload.append("\"").append("code").append("\":");
        payload.append(code).append(",");

        payload.append("\"").append("message").append("\":\"");
        payload.append(message).append("\"");

        if (devices != null) {
            payload.append(", \"").append("devices").append("\":");
            payload.append(devices);
        }
        payload.append("}");
        return payload.toString();
    }

/**********************************************************************************************
 * OTA operation
 **********************************************************************************************/
    public void onOtaPropListAddMember(String packageName, String version, String diffBaseVersion) {
        Map<String, String> member = new HashMap<>();
        member.put("package-name", packageName);
        member.put("version", version);
        member.put("diff-base-version", diffBaseVersion);

        this.otaPropList.add(member);
    }

/**********************************************************************************************
 *
 **********************************************************************************************/
    public void activate() throws JoylinkException {
    }

    public void deactivate() throws JoylinkException {
        commandManager.onDeactivate();
    }

    public boolean isCommandReady(Command command) {
        return true;
    }

/**********************************************************************************************
 * Property Message Send
 **********************************************************************************************/
    protected void sendPropertiesPost(long reportVersion, String jsonProp) {
    }

    protected void sendPropertiesPost(ServiceKey serviceKey, long reportVersion, String jsonProp) {
        if (jsonProp == null) {
            LOGGER.warning("Input jsonProp is null");
            return;
        }

        StringBuilder payload = new StringBuilder("{");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_DEV_ID).append("\":\"");
        payload.append(this.deviceId).append("\",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_TIME_STAMP).append("\":");
        payload.append(System.currentTimeMillis()/1000).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_VERSION).append("\":");
        payload.append(reportVersion).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_PROPERTIES).append("\":");
        payload.append(jsonProp).append("}");

        JoylinkQos qos = JoylinkQos.valueOf(JoylinkConfig.DEVICE_PROP_POST_QOS);
        JoylinkDeviceReportMessage message = new JoylinkDeviceReportMessage(null, qos, reportVersion,
                payload.toString(), this);
        if (enableVersioning && reportVersion != localVersion.get()) {
            LOGGER.warning("Local version number has changed, skip reporting for this round");
            return;
        }

        try {
            commandManager.runCommand(serviceKey, Command.PROP_POST, null, message);
        } catch (JoylinkTimeoutException e) {
            // async command, shouldn't receive timeout exception
        } catch (JoylinkException e) {
            LOGGER.log(Level.WARNING, "Failed to publish device report message", e);
        }
    }

    public void sendPropertiesSetResponse(String messageId, String jsonProp, long errCode) {
    
    }

    public void sendPropertiesSetResponse(ServiceKey serviceKey, String messageId, String jsonProp, long errCode) {
        long reportVersion = localVersion.get();

        StringBuilder payload = new StringBuilder("{");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_DEV_ID).append("\":\"");
        payload.append(this.deviceId).append("\",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_TIME_STAMP).append("\":");
        payload.append(System.currentTimeMillis()/1000).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_CODE).append("\":");
        payload.append(errCode).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_VERSION).append("\":");
        payload.append(reportVersion).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_PROPERTIES).append("\":");
        payload.append(jsonProp).append("}");

        JoylinkQos qos = JoylinkQos.valueOf(JoylinkConfig.DEVICE_PROP_SET_RES_QOS);
        JoylinkDeviceReportMessage message = new JoylinkDeviceReportMessage(null, qos,
                reportVersion, payload.toString(), this);

        if (enableVersioning && reportVersion != localVersion.get()) {
            LOGGER.warning("Local version number has changed, skip reporting for this round");
            return;
        }

        try {
            commandManager.runCommandResponse(serviceKey, Command.PROP_SET, message, messageId);
        } catch (JoylinkTimeoutException e) {
            // async command, shouldn't receive timeout exception
        } catch (JoylinkException e) {
            LOGGER.log(Level.WARNING, "Failed to publish device report message", e);
        }
    }

    public void sendPropertiesGetResponse(String messageId, String jsonProp, long errCode) {
    }

    public void sendPropertiesGetResponse(ServiceKey serviceKey, String messageId, String jsonProp, long errCode) {
        long reportVersion = localVersion.get();

        StringBuilder payload = new StringBuilder("{");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_DEV_ID).append("\":\"");
        payload.append(this.deviceId).append("\",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_TIME_STAMP).append("\":");
        payload.append(System.currentTimeMillis()/1000).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_CODE).append("\":");
        payload.append(errCode).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_VERSION).append("\":");
        payload.append(reportVersion).append(",");

        payload.append("\"").append(JoylinkConstant.PAYLOAD_PROPERTIES).append("\":");
        payload.append(jsonProp).append("}");

        JoylinkQos qos = JoylinkQos.valueOf(JoylinkConfig.DEVICE_PROP_GET_RES_QOS);
        JoylinkDeviceReportMessage message = new JoylinkDeviceReportMessage(null, qos,
                reportVersion, payload.toString(), this);
        if (enableVersioning && reportVersion != localVersion.get()) {
            LOGGER.warning("Local version number has changed, skip reporting for this round");
            return;
        }

        try {
            commandManager.runCommandResponse(serviceKey, Command.PROP_GET, message, messageId);
        } catch (JoylinkTimeoutException e) {
            // async command, shouldn't receive timeout exception
        } catch (JoylinkException e) {
            LOGGER.log(Level.WARNING, "Failed to publish device report message", e);
        }
    }

/**********************************************************************************************
 * Events Post Message Send
 **********************************************************************************************/
    public void sendEventPost(ServiceKey serviceKey, String deviceId, String eventKey, String paraJson) {
        if ((eventKey == null) || (paraJson == null)) {
            LOGGER.severe("[ERROR] Send Event is Failed For Information is not completed!");
            return;
        }

        // Build Event
        StringBuilder event = new  StringBuilder("{");
        event.append("\"key\":\"").append(eventKey).append("\",");
        event.append("\"parameters\":").append(paraJson).append("}");

        // Build Payload
        StringBuilder payload = new StringBuilder("{");
        payload.append("\"deviceId\":\"").append(deviceId).append("\",");
        payload.append("\"timestamp\":").append(System.currentTimeMillis()/1000).append(",");
        payload.append("\"events\":[").append(event.toString()).append("]}");

        JoylinkQos qos = JoylinkQos.valueOf(JoylinkConfig.DEVICE_EVENT_POST_QOS);
        JoylinkMessage message = new JoylinkMessage(null, qos, payload.toString());

        try {
            commandManager.runCommand(serviceKey, Command.EVT_POST, null, message);
        } catch (JoylinkTimeoutException e) {
            // async command, shouldn't receive timeout exception
        } catch (JoylinkException e) {
            LOGGER.log(Level.WARNING, "Failed to publish device report message", e);
        }
    }
    public void sendEventPost(ServiceKey serviceKey, String eventKey, String paraJson) {
        sendEventPost(serviceKey, this.deviceId, eventKey, paraJson);
    }

/**********************************************************************************************
 * Function Call Message Send
 **********************************************************************************************/
    public void sendFunctionCallResponse(String messageId, String funcKey, String outJson, long errCode) {
    }

    public void sendFunctionCallResponse(ServiceKey serviceKey, String messageId, String funcKey, String outJson, long errCode) {
        if ((funcKey == null)) {
            LOGGER.severe("[ERROR] Send Function Call is Failed: Function Key is null!");
            return;
        }

        // Build Event
        StringBuilder function = new  StringBuilder("{");
        if (outJson.length() != 0) {
            function.append("\"key\":\"").append(funcKey).append("\",");
            function.append("\"out\":").append(outJson).append("}");
        } else {
            function.append("\"key\":\"").append(funcKey).append("\"}");
        }

        // Build Payload
        StringBuilder payload = new StringBuilder("{");
        payload.append("\"deviceId\":\"").append(this.deviceId).append("\",");
        payload.append("\"timestamp\":").append(System.currentTimeMillis()/1000).append(",");
        payload.append("\"code\":").append(errCode).append(",");
        payload.append("\"functions\":[").append(function.toString()).append("]}");

        JoylinkQos qos = JoylinkQos.valueOf(JoylinkConfig.DEVICE_FUNC_CALL_RES_QOS);
        JoylinkMessage message = new JoylinkMessage(null, qos, payload.toString());

        try {
            commandManager.runCommandResponse(serviceKey, Command.FUNC_CALL, message, messageId);
        } catch (JoylinkTimeoutException e) {
            // async command, shouldn't receive timeout exception
        } catch (JoylinkException e) {
            LOGGER.log(Level.WARNING, "Failed to publish device report message", e);
        }
    }

/**********************************************************************************************
 * OTA Message Send
 **********************************************************************************************/
    public void sendOtaEventVersionChanged(String version) {
        String paraJson;
        Map<String, Object> paraMap = new HashMap<>();
        paraMap.put("version", version);
        paraMap.put("object-id", this.deviceId);
        try {
            ObjectMapper mapper = new ObjectMapper();
            paraJson = mapper.writeValueAsString(paraMap);
        } catch (IOException e) {
            LOGGER.severe("Build ota.version-cahnged event parameters fialed");;
            return;
        }

        sendEventPost(ServiceKey.OTA, "ota.ota-version-changed", paraJson);
    }

    public void sendOtaEventStateChanged(String taskId, String state, String eCode, int progress) {
        String paraJson = onOtaBuildState(taskId, state, eCode, progress);

        sendEventPost(ServiceKey.OTA, "ota.ota-state-changed", paraJson);
    }

/**********************************************************************************************
 * OTA Service Properties and Functions Call
 **********************************************************************************************/
    public String onOtaBuildState(String taskId, String state, String eCode, int progress) {
        Map<String, Object> paraMap = new HashMap<>();
        paraMap.put("task-id", taskId);
        paraMap.put("ota-state", state);
        paraMap.put("error-code", eCode);
        paraMap.put("progress", progress);
        try {
            ObjectMapper mapper = new ObjectMapper();
            return mapper.writeValueAsString(paraMap);
        } catch (IOException e) {
            LOGGER.severe("Build ota state fialed");;
            return null;
        }
    }

 /**********************************************************************************************
 * Annotation Class Method
 **********************************************************************************************/
    private Map<String, Field> getDeviceProperties(boolean enableReport, boolean allowUpdate) {
        Map<String, Field> properties = new HashMap<>();

        for (Field field : this.getClass().getDeclaredFields()) {
            JoylinkDeviceProperty annotation = field.getAnnotation(JoylinkDeviceProperty.class);
            if (annotation == null) {
                continue;
            }

            String propertyName = annotation.name().length() > 0 ? annotation.name() : field.getName();
            if ((enableReport && annotation.enableReport()) || (allowUpdate && annotation.allowUpdate())) {
                properties.put(propertyName, field);
            }
        }

        return properties;
    }
    private Map<String, Field> getDeviceProperties() {
        Map<String, Field> properties = new HashMap<>();

        for (Field field : this.getClass().getDeclaredFields()) {
            JoylinkDeviceProperty annotation = field.getAnnotation(JoylinkDeviceProperty.class);
            if (annotation == null) {
                continue;
            }

            String propertyName = annotation.name().length() > 0 ? annotation.name() : field.getName();
            properties.put(propertyName, field);
        }

        return properties;
    }

    private Map<String, Field> getDeviceFunctions() {
        Map<String, Field> functions = new HashMap<>();

        for (Field field : this.getClass().getDeclaredFields()) {
            JoylinkDeviceFunction annotation = field.getAnnotation(JoylinkDeviceFunction.class);
            if (annotation == null) {
                continue;
            }

            String functionName = annotation.name().length() > 0 ? annotation.name() : field.getName();
            functions.put(functionName, field);

        }

        return functions;
    }

    private static void invokeSetterMethod(Object target, String name, Class<?> type, Object value) throws IOException {
        String setter = "set" + Character.toUpperCase(name.charAt(0)) + name.substring(1);

        Method method;
        try {
            method = target.getClass().getMethod(setter, type);
        } catch (NoSuchMethodException | SecurityException e) {
            LOGGER.severe("Get Method setter for " + name + " failed!");
            throw new IllegalArgumentException(e);
        }

        try {
            method.invoke(target, value);
        } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
            throw new IOException(e);
        }
    }

    private static Object invokeGetterMethod(Object target, Field field) throws IOException {
        String fieldName = Character.toUpperCase(field.getName().charAt(0)) + field.getName().substring(1);
        String getter = "get" + fieldName;

        Method method;
        try {
            method = target.getClass().getMethod(getter);
        } catch (NoSuchMethodException | SecurityException e) {
            LOGGER.severe("Get Method getter for " + fieldName + " failed!");
            if (e instanceof NoSuchMethodException && boolean.class.equals(field.getType())) {
                getter = "is" + fieldName;
                try {
                    method = target.getClass().getMethod(getter);
                } catch (NoSuchMethodException | SecurityException ie) {
                    throw new IllegalArgumentException(ie);
                }
            } else {
                throw new IllegalArgumentException(e);
            }
        }

        Object value;
        try {
            value = method.invoke(target);
        } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
            throw new IOException(e);
        }
        return value;
    }

}
