/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.protocol;

import java.util.logging.Logger;

import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkTimeoutException;
import com.joylink5.services.iot.client.core.JoylinkCompletion;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.Command;

import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;

/**
 * 
 * @see com.joylink5.services.iot.client.core.JoylinkCompletion
 */
@Getter
@Setter
public class JoylinkDeviceCommand extends JoylinkCompletion {

    private static final Logger LOGGER = Logger.getLogger(JoylinkDeviceCommand.class.getName());

    private final JoylinkDeviceCommandManager commandManager;
    private final Command command;
    private final String commandId;

    private JoylinkMessage response;

    @Setter(AccessLevel.NONE)
    private Boolean requestSent;

    public JoylinkDeviceCommand(JoylinkDeviceCommandManager commandManager, Command command, String commandId,
            JoylinkMessage request, long commandTimeout, boolean isAsync) {
        super(request, commandTimeout, isAsync);
        this.commandManager = commandManager;
        this.command = command;
        this.commandId = commandId;
        this.requestSent = false;
    }

    public void put(AbstractJoylinkDevice device) throws JoylinkException {
        if (device.isCommandReady(command)) {
            _put(device);
        } else {
            LOGGER.info("Request is pending: " + command.name() + "/" + commandId);
        }
    }

    public String get(AbstractJoylinkDevice device) throws JoylinkException, JoylinkTimeoutException {
        super.get(device.getClient());
        return (response != null) ? response.getStringPayload() : null;
    }

    public boolean onReady(AbstractJoylinkDevice device) {
        try {
            LOGGER.info("Request is resumed: " + command.name() + "/" + commandId);
            _put(device);
            return true;
        } catch (JoylinkException e) {
            return false;
        }
    }

    @Override
    public void onSuccess() {
        // first callback is for the command ack, which we ignore
        if (response == null) {
            return;
        } else {
            request.setPayload(response.getPayload());
        }

        super.onSuccess();
    }

    @Override
    public void onFailure() {
        super.onFailure();
    }

    @Override
    public void onTimeout() {
        commandManager.onCommandTimeout(this);
        super.onTimeout();
    }

    private void _put(AbstractJoylinkDevice device) throws JoylinkException {
        synchronized (this) {
            if (requestSent) {
                LOGGER.warning("Request was already sent: " + command.name() + "/" + commandId);
                return;
            } else {
                requestSent = true;
            }
        }
        LOGGER.info("\r\nJoylink Publish Message:" + this.topic + "\r\n" + this.getStringPayload() + "\r\n");
        device.getClient().publish(this, timeout);
    }

}
