/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.protocol;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.joylink5.services.iot.client.JoylinkDeviceErrorCode;
import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkTimeoutException;
import com.joylink5.services.iot.client.JoylinkConstant;

import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.ServiceKey;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.Command;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.CommandAck;

import com.joylink5.services.iot.client.core.JoylinkRuntimeException;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;

import lombok.Getter;
import lombok.Setter;

/**
 * This class manages the commands sent to the platform.
 * 
 */
@Getter
@Setter
public class JoylinkDeviceCommandManager {

    private static final Logger LOGGER = Logger.getLogger(JoylinkDeviceCommandManager.class.getName());

    private static final String COMMAND_ID_FIELD = JoylinkConstant.PAYLOAD_MSG_ID;
    private static final String ERROR_CODE_FIELD = JoylinkConstant.PAYLOAD_CODE;
    private static final String ERROR_MESSAGE_FIELD = JoylinkConstant.PAYLOAD_MSG;

    private final ConcurrentMap<String, JoylinkDeviceCommand> pendingCommands;
    private final AbstractJoylinkDevice device;
    private final ObjectMapper objectMapper;


    public JoylinkDeviceCommandManager(AbstractJoylinkDevice device) {
        this.pendingCommands = new ConcurrentHashMap<>();
        this.device = device;
        this.objectMapper = new ObjectMapper();
    }

    public String runCommandSync(Command command, JoylinkMessage request) throws JoylinkException {
        try {
            return runCommand(command, request, 0, false);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because timeout is 0
            throw new JoylinkRuntimeException(e);
        }
    }

    public String runCommandSync(Command command, JoylinkMessage request, long commandTimeout) throws JoylinkException,
            JoylinkTimeoutException {
        return runCommand(command, request, commandTimeout, false);
    }

    public String runCommand(Command command, JoylinkMessage request, long commandTimeout) throws JoylinkException {
        try {
            return runCommand(command, request, commandTimeout, true);
        } catch (JoylinkTimeoutException e) {
            // We shouldn't get timeout exception because it's asynchronous call
            throw new JoylinkRuntimeException(e);
        }
    }

    public String runCommand(Command command, JoylinkMessage request, long commandTimeout, boolean isAsync)
            throws JoylinkException, JoylinkTimeoutException {
        String commandId = newCommandId();
        appendCommandId(request, commandId);

        request.setTopic(device.getClient().genTopic(null, command, null));
        JoylinkDeviceCommand deviceCommand = new JoylinkDeviceCommand(this, command, commandId, request, commandTimeout,
                isAsync);

        // pendingCommands.put(commandId, deviceCommand);
        // LOGGER.fine("Number of pending commands: " + pendingCommands.size());

        try {
            deviceCommand.put(device);
        } catch (JoylinkException e) {
            // if exception happens during publish, we remove the command
            // from the pending list as we'll never get ack for it.
            // pendingCommands.remove(commandId);
            throw e;
        }

        return deviceCommand.get(device);
    }

    public String runCommand(ServiceKey key, Command command, CommandAck ack, JoylinkMessage request)
            throws JoylinkException, JoylinkTimeoutException {
        String commandId = newCommandId();
        appendCommandId(request, commandId);

        request.setTopic(device.getClient().genTopic(key, command, ack));
        JoylinkDeviceCommand deviceCommand = new JoylinkDeviceCommand(this, command, commandId, request,
                                                        device.getClient().getServerAckTimeout(),true);

        try {
            deviceCommand.put(device);
        } catch (JoylinkException e) {
            // if exception happens during publish, we remove the command
            // from the pending list as we'll never get ack for it.
            // pendingCommands.remove(commandId);
            throw e;
        }

        return deviceCommand.get(device);
    }

    public String runCommandResponse(ServiceKey key, Command command, JoylinkMessage request, String commandId)
            throws JoylinkException, JoylinkTimeoutException {
        appendCommandId(request, commandId);

        request.setTopic(device.getClient().genTopic(key, command, CommandAck.RESPONSE));
        JoylinkDeviceCommand deviceCommand = new JoylinkDeviceCommand(this, command, commandId, request,
                                                        device.getClient().getServerAckTimeout(), true);

        // pendingCommands.put(commandId, deviceCommand);
        // LOGGER.fine("Number of pending commands: " + pendingCommands.size());

        try {
            deviceCommand.put(device);
        } catch (JoylinkException e) {
            // if exception happens during publish, we remove the command
            // from the pending list as we'll never get ack for it.
            // pendingCommands.remove(commandId);
            throw e;
        }

        return deviceCommand.get(device);
    }

    public void onCommandTimeout(JoylinkDeviceCommand command) {
        // pendingCommands.remove(command.getCommandId());
    }

    public void onDeactivate() {
        Iterator<Entry<String, JoylinkDeviceCommand>> it = pendingCommands.entrySet().iterator();
        while (it.hasNext()) {
            Entry<String, JoylinkDeviceCommand> entry = it.next();
            it.remove();

            final JoylinkDeviceCommand deviceCommand = entry.getValue();
            LOGGER.warning("Request was cancelled: " + deviceCommand.getCommand().name() + "/"
                    + deviceCommand.getCommandId());
            device.getClient().scheduleTask(new Runnable() {
                @Override
                public void run() {
                    deviceCommand.onFailure();
                }
            });
        }
    }

    private void appendCommandId(JoylinkMessage message, String commandId) throws JoylinkException {
        String payload = message.getStringPayload();
        if (payload == null) {
            payload = "{}";
        }

        try {
            JsonNode jsonNode = objectMapper.readTree(payload);
            if (!jsonNode.isObject()) {
                throw new JoylinkException("Invalid Json string in payload");
            }
            ((ObjectNode) jsonNode).put(COMMAND_ID_FIELD, commandId);

            message.setStringPayload(jsonNode.toString());
        } catch (IOException e) {
            throw new JoylinkException(e);
        }
    }

    private JoylinkDeviceCommand getPendingCommand(JoylinkMessage message) {
        String payload = message.getStringPayload();
        if (payload == null) {
            return null;
        }

        try {
            JsonNode jsonNode = objectMapper.readTree(payload);
            if (!jsonNode.isObject()) {
                return null;
            }

            JsonNode node = jsonNode.get(COMMAND_ID_FIELD);
            if (node == null) {
                return null;
            }

            String commandId = node.textValue();
            JoylinkDeviceCommand command = pendingCommands.remove(commandId);
            if (command == null) {
                return null;
            }

            node = jsonNode.get(ERROR_CODE_FIELD);
            if (node != null) {
                command.setErrorCode(JoylinkDeviceErrorCode.valueOf(node.longValue()));
            }

            node = jsonNode.get(ERROR_MESSAGE_FIELD);
            if (node != null) {
                command.setErrorMessage(node.textValue());
            }

            return command;
        } catch (IOException e) {
            return null;
        }
    }

    private String newCommandId() {
        return UUID.randomUUID().toString();
    }

}
