/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.protocol;

import java.io.IOException;
import java.util.logging.Logger;

import javax.lang.model.util.ElementScanner6;

import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkQos;
import com.joylink5.services.iot.client.JoylinkTopic;
import com.joylink5.services.iot.client.JoylinkConstant;
import com.joylink5.services.iot.client.JoylinkDeviceErrorCode;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.JsonNode;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import com.joylink5.services.iot.client.core.AbstractJoylinkClient;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.ServiceKey;

/**
 * This class extends {@link JoylinkTopic} to provide a callback function for
 * receiving the functions call.
 */
public class JoylinkDeviceFunctionsCallListener extends JoylinkTopic {

    private static final Logger LOGGER = Logger.getLogger(JoylinkDeviceFunctionsCallListener.class.getName());

    private final AbstractJoylinkClient client;

    public JoylinkDeviceFunctionsCallListener(String topic, JoylinkQos qos, AbstractJoylinkClient client) {
        super(topic, qos);
        this.client = client;
    }

    @Override
    public void onMessage(JoylinkMessage message) {
        // Error Code
        JoylinkDeviceErrorCode errCode = JoylinkDeviceErrorCode.SUCESS;

        String payload = message.getStringPayload();
        LOGGER.info("\r\nRecv Message:" + message.getTopic() + "\r\n" + payload + "\r\n");
        if (payload == null) {
            LOGGER.warning("Received Empty Message");
            errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
            return;
        }

        ServiceKey serviceKey = client.topicGetServiceKey(message.getTopic());

        ObjectMapper mapper = new ObjectMapper();
        JsonNode rootNode;
        try {
            rootNode = mapper.readTree(payload);
            if (!rootNode.isObject()) {
                throw new IOException();
            }
        } catch (IOException e) {
            LOGGER.warning("Received Invalid Payload");
            errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
            return;
        }

        String deviceId = rootNode.get(JoylinkConstant.PAYLOAD_DEV_ID).asText();

        if (deviceId != null) {
            AbstractJoylinkDevice device = client.getDevices().get(deviceId);
            if (device == null) {
                LOGGER.warning("Can not Find Device :" + deviceId);
                errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return;
            } else {
                LOGGER.info("Find Device :" + device.getDeviceId());
            }

            JsonNode node = rootNode.get(JoylinkConstant.PAYLOAD_FUNCTIONS);
            if (node == null) {
                LOGGER.warning("Missing functions Field in Payload");
                errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return;
            }
    
            String funcKey = null;
            String funInParam = null;
            if (node.isArray()) {
                for (JsonNode item : node) {
                    // System.out.println("Function Call:" + item.toString());
                    funcKey = item.get("key").asText();
                    funInParam = item.get("in").toString();
                    break;
                }
            }
            if (funcKey == null) {
                LOGGER.warning("Function Key is null!");
                errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return;
            }

            // Do Execute Function Call
            String outParam;
            device.setErrorCode(JoylinkDeviceErrorCode.SUCESS);
            if ((serviceKey!= null) && (serviceKey.equals(ServiceKey.OTA))) {
                outParam = device.onOtaFunctionsCall(funcKey, funInParam);
            } else {
                if ((serviceKey!= null) && (serviceKey.equals(ServiceKey.CONN_AGENT)) 
                    && device.getClient().getClientId().equals(device.getDeviceId())) {
                    outParam = device.onConnAgentFunctionsCall(funcKey, funInParam);
                } else {
                    outParam = device.onFunctionsCall(funcKey, funInParam);
                }
            }
            errCode = device.getErrorCode();
            device.sendFunctionCallResponse(serviceKey, rootNode.get(JoylinkConstant.PAYLOAD_MSG_ID).asText(),
                            funcKey, outParam, errCode.getValue());
        }
    }

    @Override
    public void onSuccess() {
    }

    @Override
    public void onFailure() {
        LOGGER.warning("Failed to subscribe to device topic " + topic);
    }

    @Override
    public void onTimeout() {
        LOGGER.warning("Timeout when subscribing to device topic " + topic);
    }

}
