/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.protocol;

import java.io.IOException;
import java.util.logging.Logger;

import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkQos;
import com.joylink5.services.iot.client.JoylinkTopic;
import com.joylink5.services.iot.client.JoylinkConstant;
import com.joylink5.services.iot.client.JoylinkDeviceErrorCode;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.JsonNode;

import com.joylink5.services.iot.client.core.AbstractJoylinkClient;
import com.joylink5.services.iot.client.protocol.JoylinkDeviceTopicManager.ServiceKey;

/**
 * This class extends {@link JoylinkTopic} to provide a callback function for
 * receiving the properties set.
 */
public class JoylinkDevicePropertiesSetListener extends JoylinkTopic {

    private static final Logger LOGGER = Logger.getLogger(JoylinkDevicePropertiesSetListener.class.getName());

    private final AbstractJoylinkClient client;

    public JoylinkDevicePropertiesSetListener(String topic, JoylinkQos qos, AbstractJoylinkClient client) {
        super(topic, qos);
        this.client = client;
    }

    @Override
    public void onMessage(JoylinkMessage message) {
        // Error Code
        JoylinkDeviceErrorCode errCode = JoylinkDeviceErrorCode.SUCESS;

        String payload = message.getStringPayload();
        LOGGER.info("\r\nRecv Message:" + message.getTopic() + "\r\n" + payload + "\r\n");
        if (payload == null) {
            LOGGER.warning("Received Empty Message for Topic:" + message.getTopic());
            errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
            return;
        }

        ServiceKey serviceKey = client.topicGetServiceKey(message.getTopic());

        ObjectMapper mapper = new ObjectMapper();
        JsonNode rootNode;
        try {
            rootNode = mapper.readTree(payload);
            if (!rootNode.isObject()) {
                throw new IOException();
            }
        } catch (IOException e) {
            LOGGER.warning("Received Invalid Payload !");
            errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
            return;
        }

        String deviceId = rootNode.get(JoylinkConstant.PAYLOAD_DEV_ID).asText();

        if (deviceId != null) {
            AbstractJoylinkDevice device = client.getDevices().get(deviceId);
            if (device == null) {
                LOGGER.warning("Can not Find Device :" + deviceId);
                errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return;
            } else {
                LOGGER.info("Find Device :" + device.getDeviceId());
            }

            JsonNode node = rootNode.get(JoylinkConstant.PAYLOAD_VERSION);
            if (node == null) {
                LOGGER.warning("Missing version Field in Payload");
                errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return;
            }

            long receivedVersion = node.longValue();
            if (device.enableVersioning) {
                long localVersion = device.getLocalVersion().get();
                if (receivedVersion < localVersion) {
                    LOGGER.warning("An old version of delta received for " + device.getDeviceId() + ", local "
                            + localVersion + ", received " + receivedVersion);

                    errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                    return;
                }
            }

            node = rootNode.get(JoylinkConstant.PAYLOAD_PROPERTIES);
            if (node == null) {
                LOGGER.warning("Missing properties Field in Payload");
                errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
                return;
            }

            //
            if ((serviceKey!= null) && (serviceKey.equals(ServiceKey.OTA))) {
                LOGGER.warning("OTA Properties Set Do not Support Yet!!");
                errCode = JoylinkDeviceErrorCode.BAD_REQUEST;
            } else {
                device.setErrorCode(JoylinkDeviceErrorCode.SUCESS);
                String propJson = device.onPropertiesSet(node.toString());

                errCode = device.getErrorCode();
                if (errCode.equals(JoylinkDeviceErrorCode.SUCESS)) {
                    // update local property version
                    device.getLocalVersion().set(receivedVersion);
                    LOGGER.info("Local version number updated to " + receivedVersion);
                }
                device.sendPropertiesSetResponse(rootNode.get(JoylinkConstant.PAYLOAD_MSG_ID).asText(), propJson, errCode.getValue());
            }
        }
    }

    @Override
    public void onSuccess() {
    }

    @Override
    public void onFailure() {
        LOGGER.warning("Failed to subscribe to device topic " + topic);
    }

    @Override
    public void onTimeout() {
        LOGGER.warning("Timeout when subscribing to device topic " + topic);
    }

}
