/*
 * Copyright 2020 JD.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  https://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.joylink5.services.iot.client.protocol;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.joylink5.services.iot.client.JoylinkDeviceErrorCode;
import com.joylink5.services.iot.client.JoylinkException;
import com.joylink5.services.iot.client.JoylinkMessage;
import com.joylink5.services.iot.client.JoylinkTimeoutException;
import com.joylink5.services.iot.client.core.JoylinkRuntimeException;
import com.joylink5.services.iot.client.core.AbstractJoylinkClient;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;

import lombok.Getter;
import lombok.Setter;

/**
 * This class manages the topics of the protocol.
 */
@Getter
@Setter
public class JoylinkDeviceTopicManager {

    private static final Logger LOGGER = Logger.getLogger(JoylinkDeviceTopicManager.class.getName());

    private static final String TOPIC_PREFIX = "$iot/v1/device/?";
    private static final Map<Command, String> COMMAND_PATHS;
    private static final Map<CommandAck, String> COMMAND_ACK_PATHS;
    private static final Map<ServiceKey, String> SERVICE_KEY_PATHS;

    private final AbstractJoylinkClient client;

    public static enum ServiceKey {
        CONN_AGENT, OTA, WILDCARD,
    }

    public static enum Command {
        THMD_POST,
        PROP_POST,
        PROP_SET,
        PROP_GET,
        FUNC_CALL,
        EVT_POST,
        NTP_REQ,
        NTP_RES,
    }

    public static enum CommandAck {
        RESPONSE
    }

    static {
        SERVICE_KEY_PATHS = new HashMap<ServiceKey, String>();
        SERVICE_KEY_PATHS.put(ServiceKey.CONN_AGENT, "/connection_agent");
        SERVICE_KEY_PATHS.put(ServiceKey.OTA, "/ota");
        SERVICE_KEY_PATHS.put(ServiceKey.WILDCARD, "/+");

        COMMAND_PATHS = new HashMap<Command, String>();
        COMMAND_PATHS.put(Command.THMD_POST, "/thing-model/post");
        COMMAND_PATHS.put(Command.PROP_POST, "/properties/post");
        COMMAND_PATHS.put(Command.PROP_SET, "/properties/set");
        COMMAND_PATHS.put(Command.PROP_GET, "/properties/get");
        COMMAND_PATHS.put(Command.FUNC_CALL, "/functions/call");
        COMMAND_PATHS.put(Command.EVT_POST, "/events/post");
        COMMAND_PATHS.put(Command.NTP_REQ, "/ntp/request");
        COMMAND_PATHS.put(Command.NTP_RES, "/ntp/response");


        COMMAND_ACK_PATHS = new HashMap<CommandAck, String>();
        COMMAND_ACK_PATHS.put(CommandAck.RESPONSE, "/response");
    }

    public JoylinkDeviceTopicManager(AbstractJoylinkClient client) {
        this.client = client;
    }

    public String getTopic(ServiceKey serviceKey, Command command, CommandAck ack) {
        String topic = TOPIC_PREFIX.replace("?", client.getClientId());

        if (SERVICE_KEY_PATHS.containsKey(serviceKey)) {
            topic += SERVICE_KEY_PATHS.get(serviceKey);
        }

        if (COMMAND_PATHS.containsKey(command)) {
            topic += COMMAND_PATHS.get(command);
        }

        if (COMMAND_ACK_PATHS.containsKey(ack)) {
            topic += COMMAND_ACK_PATHS.get(ack);
        }

        return topic;
    }
}
